__all__ = (
    "ObservationStartDatetime64Stacker",
    "DayObsStacker",
    "DayObsMJDStacker",
    "DayObsISOStacker",
)

import numpy as np
from astropy.time import Time

from .base_stacker import BaseStacker


class ObservationStartDatetime64Stacker(BaseStacker):
    """Add the observation start time as a numpy.datetime64."""

    cols_added = ["observationStartDatetime64"]

    def __init__(
        self,
        mjd_col="observationStartMJD",
    ):
        self.mjd_col = mjd_col
        self.cols_req = [self.mjd_col]
        self.units = [None]
        self.cols_added_dtypes = ["datetime64[ns]"]

    def _run(self, sim_data, cols_present=False):
        if cols_present:
            # Column already present in data; assume it is correct and does not
            # need recalculating.
            return sim_data

        sim_data["observationStartDatetime64"] = Time(sim_data[self.mjd_col], format="mjd").datetime64

        return sim_data


def _compute_day_obs_mjd(mjd):
    day_obs_mjd = np.floor(mjd - 0.5).astype("int")
    return day_obs_mjd


def _compute_day_obs_astropy_time(mjd):
    day_obs_time = Time(_compute_day_obs_mjd(mjd), format="mjd")
    return day_obs_time


def _compute_day_obs_iso8601(mjd):
    iso_times = _compute_day_obs_astropy_time(mjd).iso

    # Work both for mjd as a scalar and a numpy array
    if isinstance(iso_times, str):
        day_obs_iso = iso_times[:10]
    else:
        day_obs_iso = np.array([d[:10] for d in iso_times])

    return day_obs_iso


def _compute_day_obs_int(mjd):
    day_obs_int = np.array([d.replace("-", "") for d in _compute_day_obs_iso8601(mjd)])

    return day_obs_int


class DayObsStacker(BaseStacker):
    """Add dayObs as, as defined by SITCOMTN-32.

    Parameters
    ----------
    mjd_col : `str`
        The column with the observatin start MJD.
    """

    cols_added = ["dayObs"]

    def __init__(self, mjd_col="observationStartMJD"):
        self.mjd_col = mjd_col
        self.cols_req = [self.mjd_col]
        self.units = ["days"]
        self.cols_added_dtypes = [int]

    def _run(self, sim_data, cols_present=False):
        if cols_present:
            # Column already present in data; assume it is correct and does not
            # need recalculating.
            return sim_data

        sim_data[self.cols_added[0]] = _compute_day_obs_int(sim_data[self.mjd_col])
        return sim_data


class DayObsMJDStacker(BaseStacker):
    """Add dayObs defined by SITCOMTN-32, as an MJD.

    Parameters
    ----------
    mjd_col : `str`
        The column with the observatin start MJD.
    """

    cols_added = ["day_obs_mjd"]

    def __init__(self, mjd_col="observationStartMJD"):
        self.mjd_col = mjd_col
        self.cols_req = [self.mjd_col]
        self.units = ["days"]
        self.cols_added_dtypes = [int]

    def _run(self, sim_data, cols_present=False):
        if cols_present:
            # Column already present in data; assume it is correct and does not
            # need recalculating.
            return sim_data

        sim_data[self.cols_added[0]] = _compute_day_obs_mjd(sim_data[self.mjd_col])
        return sim_data


class DayObsISOStacker(BaseStacker):
    """Add dayObs as defined by SITCOMTN-32, in ISO 8601 format.

    Parameters
    ----------
    mjd_col : `str`
        The column with the observatin start MJD."""

    cols_added = ["day_obs_iso8601"]

    def __init__(self, mjd_col="observationStartMJD"):
        self.mjd_col = mjd_col
        self.cols_req = [self.mjd_col]
        self.units = [None]
        self.cols_added_dtypes = [(str, 10)]

    def _run(self, sim_data, cols_present=False):
        if cols_present:
            # Column already present in data; assume it is correct and does not
            # need recalculating.
            return sim_data

        sim_data[self.cols_added[0]] = _compute_day_obs_iso8601(sim_data[self.mjd_col])
        return sim_data



__all__ = ("TwoDMap", "VisitPairsHist")

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import colors

from .perceptual_rainbow import make_pr_cmap
from .plot_handler import BasePlotter

perceptual_rainbow = make_pr_cmap()


class TwoDMap(BasePlotter):
    def __init__(self):
        self.plot_type = "TwoD"
        self.object_plotter = False
        self.default_plot_dict = {
            "title": None,
            "xlabel": None,
            "ylabel": None,
            "label": None,
            "log_scale": False,
            "cbar_format": None,
            "cbarTitle": "Count",
            "cmap": perceptual_rainbow,
            "percentile_clip": None,
            "color_min": None,
            "color_max": None,
            "zp": None,
            "norm_val": None,
            "cbar_edge": True,
            "n_ticks": None,
            "aspect": "auto",
            "xextent": None,
            "origin": None,
            "figsize": None,
        }

    def __call__(self, metric_value, slicer, user_plot_dict, fig=None):
        """
        Parameters
        ----------
        metric_value : `numpy.ma.MaskedArray`
            The metric values from the bundle.
        slicer : `rubin_sim.maf.slicers.TwoDSlicer`
            The slicer.
        user_plot_dict : `dict`
            Dictionary of plot parameters set by user
            (overrides default values).
        fig : `matplotlib.figure.Figure`
            Matplotlib figure number to use. Default = None, starts new figure.

        Returns
        -------
        fig : `matplotlib.figure.Figure`
            Figure with the plot.
        """
        if "Healpix" in slicer.slicer_name:
            self.default_plot_dict["ylabel"] = "Healpix ID"
        elif "User" in slicer.slicer_name:
            self.default_plot_dict["ylabel"] = "User Field ID"

        plot_dict = {}
        plot_dict.update(self.default_plot_dict)
        # Don't clobber with None
        for key in user_plot_dict:
            if user_plot_dict[key] is not None:
                plot_dict[key] = user_plot_dict[key]

        if plot_dict["xextent"] is None:
            plot_dict["xextent"] = [0, metric_value[0, :].size]

        if plot_dict["log_scale"]:
            norm = colors.LogNorm()
        else:
            norm = None

        # Mask out values below the color minimum so they show up as white
        if plot_dict["color_min"] is not None:
            low_vals = np.where(metric_value.data < plot_dict["color_min"])
            metric_value.mask[low_vals] = True

        if fig is None:
            fig = plt.figure(figsize=plot_dict["figsize"])
        ax = fig.add_subplot(111)
        yextent = [0, slicer.nslice - 1]
        xextent = plot_dict["xextent"]
        extent = []
        extent.extend(xextent)
        extent.extend(yextent)
        image = ax.imshow(
            metric_value,
            vmin=plot_dict["color_min"],
            vmax=plot_dict["color_max"],
            aspect=plot_dict["aspect"],
            cmap=plot_dict["cmap"],
            norm=norm,
            extent=extent,
            interpolation="none",
            origin=plot_dict["origin"],
        )
        cb = plt.colorbar(image)

        ax.set_xlabel(plot_dict["xlabel"])
        ax.set_ylabel(plot_dict["ylabel"])
        ax.set_title(plot_dict["title"])
        cb.set_label(plot_dict["cbarTitle"])

        # Fix white space on pdf's
        if plot_dict["cbar_edge"]:
            cb.solids.set_edgecolor("face")
        return fig


class VisitPairsHist(BasePlotter):
    """
    Given an TwoDSlicer, figure out what fraction of observations
    are in singles, pairs, triples, etc.


    Parameters
    ----------
    metric_value : `numpy.ma.MaskedArray`
        The metric values from the bundle.
    slicer : `rubin_sim.maf.slicers.TwoDSlicer`
        The slicer.
    user_plot_dict : `dict`
        Dictionary of plot parameters set by user
        (overrides default values).
    fig : `matplotlib.figure.Figure`
        Matplotlib figure number to use. Default = None, starts new figure.

    Returns
    -------
    fig : `matplotlib.figure.Figure`
        Figure with the plot.
    """

    def __init__(self):
        self.plot_type = "TwoD"
        self.object_plotter = False
        self.default_plot_dict = {
            "title": None,
            "xlabel": "N visits per night per field",
            "ylabel": "N Visits",
            "label": None,
            "color": "b",
            "xlim": [0, 20],
            "ylim": None,
            "figsize": None,
        }

    def __call__(self, metric_value, slicer, user_plot_dict, fig=None):
        plot_dict = {}
        plot_dict.update(self.default_plot_dict)
        # Don't clobber with None
        for key in user_plot_dict:
            if user_plot_dict[key] is not None:
                plot_dict[key] = user_plot_dict[key]

        max_val = metric_value.max()
        bins = np.arange(0.5, max_val + 0.5, 1)

        vals, bins = np.histogram(metric_value, bins)
        xvals = (bins[:-1] + bins[1:]) / 2.0

        if fig is None:
            fig = plt.figure(figsize=plot_dict["figsize"])
        ax = fig.add_subplot(111)
        ax.bar(xvals, vals * xvals, color=plot_dict["color"], label=plot_dict["label"])
        ax.set_xlabel(plot_dict["xlabel"])
        ax.set_ylabel(plot_dict["ylabel"])
        ax.set_title(plot_dict["title"])
        ax.set_xlim(plot_dict["xlim"])
        ax.set_ylim(plot_dict["ylim"])

        return fig



# cumulative one dimensional movie slicer
__all__ = ("MovieSlicer",)

import os
import subprocess
import warnings
from functools import wraps
from subprocess import CalledProcessError

import numpy as np

from rubin_sim.maf.stackers import ColInfo
from rubin_sim.maf.utils import optimal_bins

from .base_slicer import BaseSlicer


class MovieSlicer(BaseSlicer):
    def __init__(
        self,
        slice_col_name=None,
        slice_col_units=None,
        bins=None,
        bin_min=None,
        bin_max=None,
        bin_size=None,
        verbose=True,
        badval=0,
        cumulative=True,
        force_no_ffmpeg=False,
    ):
        """The 'movieSlicer' acts similarly to the OneDSlicer
        (slices on one data column).
        However, the data slices from the movieSlicer are intended to be
        fed to another slicer, which then
        (together with a set of Metrics) calculates metric values +
        plots at each slice created by the movieSlicer.
        The job of the movieSlicer is to track those slices and
        put them together into a movie.

        'slice_col_name' is the name of the data column to use for slicing.
        'slice_col_units' lets the user set the units (for plotting purposes)
        of the slice column.
        'bins' can be a numpy array with the binpoints for sliceCol
        or a single integer value
        (if a single value, this will be used as the number of bins,
        together with data min/max or bin_min/max),
        as in numpy's histogram function.
        If 'bin_size' is used, this will override the bins value and
        will be used together with the data min/max
        or bin_min/Max to set the binpoint values.

        Bins work like numpy histogram bins: the last 'bin' value is end
        value of last bin; all bins except for last bin are half-open
        ([a, b>), the last one is ([a, b]).

        The movieSlicer stitches individual frames together into a movie
        using ffmpeg. Thus, on instantiation it checks that ffmpeg is
        available and will raise and exception if not.
        This behavior can be overriden using force_no_ffmpeg = True
        (in order to create a movie later perhaps).
        """
        # Check for ffmpeg.
        if not force_no_ffmpeg:
            try:
                fnull = open(os.devnull, "w")
                subprocess.check_call(["which", "ffmpeg"], stdout=fnull)
            except CalledProcessError:
                raise Exception(
                    "Could not find ffmpeg on the system, so will not be able to create movie."
                    " Use force_no_ffmpeg=True to override this error and create individual images."
                )
        super().__init__(verbose=verbose, badval=badval)
        self.slice_col_name = slice_col_name
        self.columns_needed = [slice_col_name]
        self.bins = bins
        self.bin_min = bin_min
        self.bin_max = bin_max
        self.bin_size = bin_size
        self.cumulative = cumulative
        if slice_col_units is None:
            co = ColInfo()
            self.slice_col_units = co.get_units(self.slice_col_name)
        self.slicer_init = {
            "slice_col_name": self.slice_col_name,
            "slice_col_units": slice_col_units,
            "badval": badval,
        }

    def setup_slicer(self, sim_data, maps=None):
        """
        Set up bins in slicer.
        """
        if self.slice_col_name is None:
            raise Exception("slice_col_name was not defined when slicer instantiated.")
        slice_col = sim_data[self.slice_col_name]
        # Set bin min/max values.
        if self.bin_min is None:
            self.bin_min = slice_col.min()
        if self.bin_max is None:
            self.bin_max = slice_col.max()
        # Give warning if bin_min = bin_max, and do something at
        # least slightly reasonable.
        if self.bin_min == self.bin_max:
            warnings.warn(
                "bin_min = bin_max (maybe your data is single-valued?). "
                "Increasing bin_max by 1 (or 2*bin_size, if bin_size set)."
            )
            if self.bin_size is not None:
                self.bin_max = self.bin_max + 2 * self.bin_size
            else:
                self.bin_max = self.bin_max + 1
        # Set bins.
        # Using bin_size.
        if self.bin_size is not None:
            if self.bins is not None:
                warnings.warn(
                    "Both bin_size and bins have been set; Using bin_size %f only." % (self.bin_size)
                )
            self.bins = np.arange(
                self.bin_min,
                self.bin_max + self.bin_size / 2.0,
                float(self.bin_size),
                "float",
            )
        # Using bins value.
        else:
            # Bins was a sequence (np array or list)
            if hasattr(self.bins, "__iter__"):
                self.bins = np.sort(self.bins)
            # Or bins was a single value.
            else:
                if self.bins is None:
                    self.bins = optimal_bins(slice_col, self.bin_min, self.bin_max)
                nbins = int(self.bins)
                self.bin_size = (self.bin_max - self.bin_min) / float(nbins)
                self.bins = np.arange(
                    self.bin_min,
                    self.bin_max + self.bin_size / 2.0,
                    self.bin_size,
                    "float",
                )
        # Set shape / nbins to be one less than # of bins because
        # last binvalue is RH edge only
        self.nslice = len(self.bins) - 1
        self.shape = self.nslice
        # Set slice_point metadata.
        self.slice_points["sid"] = np.arange(self.nslice)
        self.slice_points["bins"] = self.bins
        # Add metadata from maps.
        self._run_maps(maps)
        # Set up data slicing.
        self.sim_idxs = np.argsort(sim_data[self.slice_col_name])
        simFieldsSorted = np.sort(sim_data[self.slice_col_name])
        # "left" values are location where simdata == bin value
        # Note that this setup will place any the visits beyond
        # the last bin into the last bin.
        self.left = np.searchsorted(simFieldsSorted, self.bins[:-1], "left")
        self.left = np.concatenate(
            (
                self.left,
                np.array(
                    [
                        len(self.sim_idxs),
                    ]
                ),
            )
        )
        # Set up _slice_sim_data method for this class.
        if self.cumulative:

            @wraps(self._slice_sim_data)
            def _slice_sim_data(islice):
                """
                Slice sim_data on oneD sliceCol, to
                return relevant indexes for slice_point.
                """
                # this is the important part.
                # The ids here define the pieces of data that get
                # passed on to subsequent slicers
                # cumulative version of 1D slicing
                idxs = self.sim_idxs[0 : self.left[islice + 1]]
                return {
                    "idxs": idxs,
                    "slice_point": {
                        "sid": islice,
                        "bin_left": self.bins[0],
                        "bin_right": self.bins[islice + 1],
                    },
                }

            setattr(self, "_slice_sim_data", _slice_sim_data)
        else:

            @wraps(self._slice_sim_data)
            def _slice_sim_data(islice):
                """
                Slice sim_data on oneD sliceCol,
                to return relevant indexes for slice_point.
                """
                idxs = self.sim_idxs[self.left[islice] : self.left[islice + 1]]
                return {
                    "idxs": idxs,
                    "slice_point": {
                        "sid": islice,
                        "bin_left": self.bins[islice],
                        "bin_right": self.bins[islice + 1],
                    },
                }

            setattr(self, "_slice_sim_data", _slice_sim_data)

    def __eq__(self, other_slicer):
        """
        Evaluate if slicers are equivalent.
        """
        if isinstance(other_slicer, MovieSlicer):
            return np.array_equal(other_slicer.slice_points["bins"], self.slice_points["bins"])
        else:
            return False

    def make_movie(
        self,
        outfileroot,
        sliceformat,
        plot_type,
        fig_format,
        out_dir="Output",
        ips=10.0,
        fps=10.0,
    ):
        """Takes in metric and slicer metadata and calls
        ffmpeg to stitch together output files.
        """
        if not os.path.isdir(out_dir):
            raise Exception("Cannot find output directory %s with movie input files." % (out_dir))
        # make video
        # ffmpeg -r 30 -i [image names - FilterColors_%03d_SkyMap.png] -r 30
        #    -pix_fmt yuv420p -crf 18 -preset slower outfile
        callList = [
            "ffmpeg",
            "-r",
            str(ips),
            "-i",
            os.path.join(
                out_dir,
                "%s_%s_%s.%s" % (outfileroot, sliceformat, plot_type, fig_format),
            ),
            "-vf",
            "pad=ceil(iw/2)*2:ceil(ih/2)*2",
            "-r",
            str(fps),
            "-pix_fmt",
            "yuv420p",
            "-crf",
            "18",
            "-preset",
            "slower",
            os.path.join(
                out_dir,
                "%s_%s_%s_%s.mp4" % (outfileroot, plot_type, str(ips), str(fps)),
            ),
        ]
        print("Attempting to call ffmpeg with:")
        print(" ".join(callList))
        subprocess.check_call(callList)
        # make thumbnail gif
        callList = [
            "ffmpeg",
            "-i",
            os.path.join(
                out_dir,
                "%s_%s_%s_%s.mp4" % (outfileroot, plot_type, str(ips), str(fps)),
            ),
            "-vf",
            "pad=ceil(iw/2)*2:ceil(ih/2)*2",
            "-t",
            str(10),
            "-r",
            str(10),
            os.path.join(
                out_dir,
                "%s_%s_%s_%s.gif" % (outfileroot, plot_type, str(ips), str(fps)),
            ),
        ]
        print("converting to animated gif with:")
        print(" ".join(callList))
        subprocess.check_call(callList)



# The base class for all spatial slicers.
# Slicers are 'data slicers' at heart; spatial slicers slice data by RA/Dec and
# return the relevant indices in the sim_data to the metric.
# The primary things added here are the methods to slice the data
# (for any spatial slicer) as this uses a KD-tree built on spatial
# (RA/Dec type) indexes.

__all__ = ("BaseSpatialSlicer",)

import warnings
from functools import wraps

import numpy as np
import rubin_scheduler.utils as simsUtils

from rubin_sim.maf.plots.spatial_plotters import BaseHistogram, BaseSkyMap

from .base_slicer import BaseSlicer


def rotate(x, y, rotation_angle_rad):
    """rotate 2d points around the origin (0,0)"""
    cos_rad = np.cos(rotation_angle_rad)
    sin_rad = np.sin(rotation_angle_rad)
    qx = cos_rad * x - sin_rad * y
    qy = sin_rad * x + cos_rad * y
    return qx, qy


class BaseSpatialSlicer(BaseSlicer):
    """Base spatial slicer object, contains additional functionality
    for spatial slicing, including setting up and traversing a kdtree
    containing the simulated data points.

    Parameters
    ----------
    lon_col : `str`, optional
        Name of the longitude (RA equivalent) column.
    lat_col : `str`, optional
        Name of the latitude (Dec equivalent) column.
    rot_sky_pos_col_name : `str`, optional
        Name of the rotSkyPos column in the input  data.
        Only used if use_camera is True.
        Describes the orientation of the camera orientation on the sky.
    lat_lon_deg : `bool`, optional
        Flag indicating whether lat and lon values from input data are
        in degrees (True) or radians (False).
    verbose : `bool`, optional
        Verbosity flag - noisy or quiet.
    badval : `float`, optional
        Bad value flag, relevant for plotting.
    leafsize : `int`, optional
        Leafsize value for kdtree.
    radius : `float`, optional
        Radius for matching in the kdtree.
        Equivalent to the radius of the FOV, in degrees.
    use_camera : `bool`, optional
        Flag to indicate whether to use the LSST camera footprint or not.
    camera_footprint_file : `str`, optional
        Name of the camera footprint map to use.
        Can be None, which will use the default file.
    """

    def __init__(
        self,
        lon_col="fieldRA",
        lat_col="fieldDec",
        rot_sky_pos_col_name="rotSkyPos",
        lat_lon_deg=True,
        verbose=True,
        badval=-666,
        leafsize=100,
        radius=2.45,
        use_camera=True,
        camera_footprint_file=None,
    ):
        super().__init__(verbose=verbose, badval=badval)
        self.lon_col = lon_col
        self.lat_col = lat_col
        self.latLonDeg = lat_lon_deg
        self.rotSkyPosColName = rot_sky_pos_col_name
        self.columns_needed = [lon_col, lat_col, rot_sky_pos_col_name]
        self.use_camera = use_camera
        self.camera_footprint_file = camera_footprint_file
        self.radius = radius
        self.leafsize = leafsize
        self.slicer_init = {
            "lon_col": self.lon_col,
            "lat_col": self.lat_col,
            "radius": self.radius,
            "badval": self.badval,
            "use_camera": self.use_camera,
        }
        # RA and Dec are required slice_point info for any spatial slicer.
        # slice_point RA/Dec are in radians.
        self.slice_points["sid"] = None
        self.slice_points["ra"] = None
        self.slice_points["dec"] = None
        self.nslice = None
        self.shape = None
        self.plot_funcs = [BaseHistogram, BaseSkyMap]

    def setup_slicer(self, sim_data, maps=None):
        """Use sim_data[self.lon_col] and sim_data[self.lat_col]
        (in radians) to set up KDTree.

        Parameters
        -----------
        sim_data : `numpy.ndarray`
            The simulated data, including the location of each pointing.
        maps : `list` of `rubin_sim.maf.maps` objects, optional
            List of maps (such as dust extinction) that will run to build up
            additional data at each slice_point. This additional data
            is available to metrics via the slice_point dictionary.
        """
        if maps is not None:
            if self.cache_size != 0 and len(maps) > 0:
                warnings.warn(
                    "Warning:  Loading maps but cache on." "Should probably set use_cache=False in slicer."
                )
            self._run_maps(maps)
        self._set_rad(self.radius)

        if self.latLonDeg:
            self.data_ra = np.radians(sim_data[self.lon_col])
            self.data_dec = np.radians(sim_data[self.lat_col])
            self.data_rot = np.radians(sim_data[self.rotSkyPosColName])
        else:
            self.data_ra = sim_data[self.lon_col]
            self.data_dec = sim_data[self.lat_col]
            self.data_rot = sim_data[self.rotSkyPosColName]
        if self.use_camera:
            self._setupLSSTCamera()

        self._build_tree(self.data_ra, self.data_dec, self.leafsize)

        @wraps(self._slice_sim_data)
        def _slice_sim_data(islice):
            """Return indexes for relevant opsim data at slice_point
            (slice_point=lon_col/lat_col value .. usually ra/dec).
            """

            # Build dict for slice_point info
            slice_point = {"sid": islice}
            sx, sy, sz = simsUtils._xyz_from_ra_dec(
                self.slice_points["ra"][islice], self.slice_points["dec"][islice]
            )
            # Query against tree.
            indices = self.opsimtree.query_ball_point((sx, sy, sz), self.rad)

            if (self.use_camera) & (len(indices) > 0):
                # Find the indices *of those indices*
                # which fall in the camera footprint
                camera_idx = self.camera(
                    self.slice_points["ra"][islice],
                    self.slice_points["dec"][islice],
                    self.data_ra[indices],
                    self.data_dec[indices],
                    self.data_rot[indices],
                )
                indices = np.array(indices)[camera_idx]

            # Loop through all the slice_point keys.
            # If the first dimension of slice_point[key] has the same shape
            # as the slicer, assume it is information per slice_point.
            # Otherwise, pass the whole slice_point[key] information.
            # Useful for stellar LF maps where we want to pass only the
            # relevant LF and the bins that go with it.
            for key in self.slice_points:
                if len(np.shape(self.slice_points[key])) == 0:
                    keyShape = 0
                else:
                    keyShape = np.shape(self.slice_points[key])[0]
                if keyShape == self.nslice:
                    slice_point[key] = self.slice_points[key][islice]
                else:
                    slice_point[key] = self.slice_points[key]
            return {"idxs": indices, "slice_point": slice_point}

        setattr(self, "_slice_sim_data", _slice_sim_data)

    def _setupLSSTCamera(self):
        """If we want to include the camera chip gaps, etc."""
        self.camera = simsUtils.LsstCameraFootprint(
            units="radians", footprint_file=self.camera_footprint_file
        )

    def _build_tree(self, sim_dataRa, sim_dataDec, leafsize=100):
        """Build KD tree on sim_dataRA/Dec.

        sim_dataRA, sim_dataDec = RA and Dec values (in radians).
        leafsize = the number of Ra/Dec pointings in each leaf node.
        """
        self.opsimtree = simsUtils._build_tree(sim_dataRa, sim_dataDec, leafsize)

    def _set_rad(self, radius=1.75):
        """Set radius (in degrees) for kdtree search."""
        self.rad = simsUtils.xyz_angular_radius(radius)



import unittest

import numpy as np

import rubin_sim.maf.metrics as metrics
import rubin_sim.maf.stackers as stackers
from rubin_sim.maf.metrics.base_metric import BaseMetric


class OldTeffMetric(BaseMetric):
    """
    Effective time equivalent for a given set of visits.
    """

    def __init__(
        self,
        m5_col="fiveSigmaDepth",
        filter_col="filter",
        metric_name="tEff",
        fiducial_depth=None,
        teff_base=30.0,
        normed=False,
        **kwargs,
    ):
        self.m5_col = m5_col
        self.filter_col = filter_col
        if fiducial_depth is None:
            # From reference von Karman 500nm zenith seeing of 0.69"
            # median zenith dark seeing from sims_skybrightness_pre
            # airmass = 1
            # 2 "snaps" of 15 seconds each
            # m5_flat_sed sysEngVals from rubin_sim
            #   commit 6d03bd49550972e48648503ed60784a4e6775b82 (2021-05-18)
            # These include constants from:
            #   https://github.com/lsst-pst/syseng_throughputs/blob/master/notebooks/generate_sims_values.ipynb
            #   commit 7abb90951fcbc70d9c4d0c805c55a67224f9069f (2021-05-05)
            # See https://github.com/lsst-sims/smtn-002/blob/master/notebooks/teff_fiducial.ipynb
            self.depth = {
                "u": 23.71,
                "g": 24.67,
                "r": 24.24,
                "i": 23.82,
                "z": 23.21,
                "y": 22.40,
            }
        else:
            if isinstance(fiducial_depth, dict):
                self.depth = fiducial_depth
            else:
                raise ValueError("fiducial_depth should be None or dictionary")
        self.teff_base = teff_base
        self.normed = normed
        if self.normed:
            units = ""
        else:
            units = "seconds"
        super(OldTeffMetric, self).__init__(
            col=[m5_col, filter_col], metric_name=metric_name, units=units, **kwargs
        )
        if self.normed:
            self.comment = "Normalized effective time"
        else:
            self.comment = "Effect time"
        self.comment += " of a series of observations, evaluating the equivalent amount of time"
        self.comment += " each observation would require if taken at a fiducial limiting magnitude."
        self.comment += " Fiducial depths are : %s" % self.depth
        if self.normed:
            self.comment += " Normalized by the total amount of time actual on-sky."

    def run(self, data_slice, slice_point=None):
        filters = np.unique(data_slice[self.filter_col])
        teff = 0.0
        for f in filters:
            match = np.where(data_slice[self.filter_col] == f)[0]
            teff += (10.0 ** (0.8 * (data_slice[self.m5_col][match] - self.depth[f]))).sum()
        teff *= self.teff_base
        if self.normed:
            # Normalize by the t_eff equivalent if each observation
            # was at the fiducial depth.
            teff = teff / (self.teff_base * data_slice[self.m5_col].size)
        return teff


class TestTechnicalMetrics(unittest.TestCase):
    def test_n_changes_metric(self):
        """
        Test the NChanges metric.
        """
        filters = np.array(["u", "u", "g", "g", "r"])
        visit_times = np.arange(0, filters.size, 1)
        data = np.core.records.fromarrays([visit_times, filters], names=["observationStartMJD", "filter"])
        metric = metrics.NChangesMetric()
        result = metric.run(data)
        self.assertEqual(result, 2)
        filters = np.array(["u", "g", "u", "g", "r"])
        data = np.core.records.fromarrays([visit_times, filters], names=["observationStartMJD", "filter"])
        metric = metrics.NChangesMetric()
        result = metric.run(data)
        self.assertEqual(result, 4)

    def test_min_time_between_states_metric(self):
        """
        Test the minTimeBetweenStates metric.
        """
        filters = np.array(["u", "g", "g", "r"])
        visit_times = np.array([0, 5, 6, 7])  # days
        data = np.core.records.fromarrays([visit_times, filters], names=["observationStartMJD", "filter"])
        metric = metrics.MinTimeBetweenStatesMetric()
        result = metric.run(data)  # minutes
        self.assertEqual(result, 2 * 24.0 * 60.0)
        data["filter"] = np.array(["u", "u", "u", "u"])
        result = metric.run(data)
        self.assertEqual(result, metric.badval)

    def test_n_state_changes_faster_than_metric(self):
        """
        Test the NStateChangesFasterThan metric.
        """
        filters = np.array(["u", "g", "g", "r"])
        visit_times = np.array([0, 5, 6, 7])  # days
        data = np.core.records.fromarrays([visit_times, filters], names=["observationStartMJD", "filter"])
        metric = metrics.NStateChangesFasterThanMetric(cutoff=3 * 24 * 60)
        result = metric.run(data)  # minutes
        self.assertEqual(result, 1)

    def test_max_state_changes_within_metric(self):
        """
        Test the MaxStateChangesWithin metric.
        """
        filters = np.array(["u", "g", "r", "u", "g", "r"])
        visit_times = np.array([0, 1, 1, 4, 6, 7])  # days
        data = np.core.records.fromarrays([visit_times, filters], names=["observationStartMJD", "filter"])
        metric = metrics.MaxStateChangesWithinMetric(timespan=1 * 24 * 60)
        result = metric.run(data)  # minutes
        self.assertEqual(result, 2)
        filters = np.array(["u", "g", "g", "u", "g", "r", "g", "r"])
        visit_times = np.array([0, 1, 1, 4, 4, 7, 8, 8])  # days
        data = np.core.records.fromarrays([visit_times, filters], names=["observationStartMJD", "filter"])
        metric = metrics.MaxStateChangesWithinMetric(timespan=1 * 24 * 60)
        result = metric.run(data)  # minutes
        self.assertEqual(result, 3)

        filters = np.array(["u", "g"])
        visit_times = np.array([0, 1])  # days
        data = np.core.records.fromarrays([visit_times, filters], names=["observationStartMJD", "filter"])
        metric = metrics.MaxStateChangesWithinMetric(timespan=1 * 24 * 60)
        result = metric.run(data)  # minutes
        self.assertEqual(result, 1)

        filters = np.array(["u", "u"])
        visit_times = np.array([0, 1])  # days
        data = np.core.records.fromarrays([visit_times, filters], names=["observationStartMJD", "filter"])
        metric = metrics.MaxStateChangesWithinMetric(timespan=1 * 24 * 60)
        result = metric.run(data)  # minutes
        self.assertEqual(result, 0)

    def test_teff_regression(self):
        """Test this teff implementation matches the old one."""
        num_points = 50
        bands = tuple("ugrizy")
        rng = np.random.default_rng(seed=6563)

        m5 = 24 + rng.random(num_points)
        filters = rng.choice(bands, num_points)
        fiducial_depth = {b: 24 + rng.random() for b in bands}
        exposure_time = np.full(num_points, 30.0, dtype=float)
        data = np.core.records.fromarrays(
            [m5, filters, exposure_time], names=["fiveSigmaDepth", "filter", "visitExposureTime"]
        )
        teff_stacker = stackers.TeffStacker(fiducial_depth=fiducial_depth, teff_base=30.0)
        data = teff_stacker.run(data)

        metric = metrics.SumMetric(col="t_eff")
        result = metric.run(data)
        old_metric = OldTeffMetric(fiducial_depth=fiducial_depth, teff_base=30.0)
        old_result = old_metric.run(data)
        self.assertEqual(result, old_result)

        data = np.core.records.fromarrays(
            [m5, filters, exposure_time], names=["fiveSigmaDepth", "filter", "visitExposureTime"]
        )
        teff_stacker = stackers.TeffStacker(fiducial_depth=fiducial_depth, teff_base=30.0, normed=True)
        data = teff_stacker.run(data)
        metric = metrics.MeanMetric(col="t_eff")
        result = metric.run(data)
        old_metric = OldTeffMetric(fiducial_depth=fiducial_depth, teff_base=30.0, normed=True)
        old_result = old_metric.run(data)
        self.assertAlmostEqual(result, old_result)

    def test_open_shutter_fraction_metric(self):
        """
        Test the open shutter fraction metric.
        """
        nvisit = 10
        exptime = 30.0
        slewtime = 30.0
        visit_exp_time = np.ones(nvisit, dtype="float") * exptime
        visit_time = np.ones(nvisit, dtype="float") * (exptime + 0.0)
        slew_time = np.ones(nvisit, dtype="float") * slewtime
        data = np.core.records.fromarrays(
            [visit_exp_time, visit_time, slew_time],
            names=["visitExposureTime", "visitTime", "slewTime"],
        )
        metric = metrics.OpenShutterFractionMetric()
        result = metric.run(data)
        self.assertEqual(result, 0.5)

    def test_brute_osf_metric(self):
        """
        Test the open shutter fraction metric.
        """
        nvisit = 10
        exptime = 30.0
        slewtime = 30.0
        visit_exp_time = np.ones(nvisit, dtype="float") * exptime
        visit_time = np.ones(nvisit, dtype="float") * (exptime + 0.0)
        slew_time = np.ones(nvisit, dtype="float") * slewtime
        mjd = np.zeros(nvisit) + np.add.accumulate(visit_exp_time) + np.add.accumulate(slew_time)
        mjd = mjd / 60.0 / 60.0 / 24.0
        data = np.core.records.fromarrays(
            [visit_exp_time, visit_time, slew_time, mjd],
            names=[
                "visitExposureTime",
                "visit_time",
                "slew_time",
                "observationStartMJD",
            ],
        )
        metric = metrics.BruteOSFMetric()
        result = metric.run(data)
        self.assertGreater(result, 0.5)
        self.assertLess(result, 0.6)


if __name__ == "__main__":
    unittest.main()



import os
import sys

import h5py
import healpy as hp
import numpy as np
import rubin_scheduler.utils as utils
from astropy.coordinates import AltAz, EarthLocation, get_sun
from astropy.time import Time

import rubin_sim.skybrightness as sb


def generate_sky(
    mjd0=59560.2,
    mjd_max=59565.2,
    timestep=5.0,
    timestep_max=15.0,
    outfile=None,
    outpath=None,
    nside=32,
    sunLimit=-11.5,
    airmass_overhead=1.5,
    dm=0.2,
    airmass_limit=2.5,
    alt_limit=86.5,
    verbose=True,
):
    """
    Pre-compute the sky brighntess for a series of mjd dates at the LSST site.

    Parameters
    ----------
    mjd0 : `float`
        The starting MJD time
    duration : `float`
        The length of time to generate sky maps for (years)
    timestep : `float`
        The timestep between sky maps (minutes)
    timestep_max : `float`
        The maximum allowable timestep (minutes)
    outfile : `str`
        The name of the output file to save the results in
    nside : `int`
        The nside to create the healpixel maps.
        Default of 32 matches expectation of rubin_scheduler.
    sunLimit : `float`
        The maximum altitude of the sun to try and generate maps for.
        MJDs with a higher sun altitude are dropped.
    fieldID : `bool`
        If True, computes sky magnitudes at OpSim field locations.
        If False computes at healpixel centers.
    airmass_overhead : `float`
        The airmass region to demand sky models are well matched before
        dropping and assuming the timestep can be interpolated.
    dm : `float`
        If a skymap can be interpolated from neighboring maps with
        precision dm, that mjd is dropped.
    airmass_limit : `float`
        Pixels with an airmass greater than airmass_limit are masked.
    moon_dist_limit : float
        Pixels (fields) closer than moon_dist_limit (degrees) are masked.
    planet_dist_limit : `float`
        Pixels (fields) closer than planet_dist_limit (degrees) to Venus,
        Mars, Jupiter, or Saturn are masked.
    alt_limit : `float`
        Altitude limit of the telescope (degrees).
        Altitudes higher than this are masked.

    Returns
    -------
    dict_of_lists : `dict`
        includes key-value pairs:
        mjds : the MJD at every computation. Not necessarily evenly spaced.
        airmass : the airmass maps for each MJD
        masks : The `bool` mask map for each MJD
        (True means the pixel should be masked)
        sunAlts : The sun altitude at each MJD
    sky_brightness : `dict`
        Has keys for each u,g,r,i,z,y filter.
        Each one is a 2-d array with dimensions of healpix ID and
        mjd (matched to the mjd list above).
    """

    sunLimit_rad = np.radians(sunLimit)

    # Set the time steps
    timestep = timestep / 60.0 / 24.0  # Convert to days
    timestep_max = timestep_max / 60.0 / 24.0  # Convert to days
    # Switch the indexing to opsim field ID if requested

    # Look at the mjds and toss ones where the sun is up
    mjds = np.arange(mjd0, mjd_max + timestep, timestep)
    sunAlts = np.zeros(mjds.size, dtype=float)

    if outfile is None:
        outfile = "%i_%i.h5" % (mjd0, mjd_max)
    if outpath is not None:
        outfile = os.path.join(outpath, outfile)

    site = utils.Site("LSST")

    location = EarthLocation(lat=site.latitude, lon=site.longitude, height=site.height)
    t_sparse = Time(mjds, format="mjd", location=location)

    sun = get_sun(t_sparse)
    aa = AltAz(location=location, obstime=t_sparse)
    sunAlts = sun.transform_to(aa)

    mjds = mjds[np.where(sunAlts.alt.rad <= sunLimit_rad)]
    required_mjds = mjds[::3]

    hpindx = np.arange(hp.nside2npix(nside))
    ra, dec = utils.hpid2_ra_dec(nside, hpindx)

    if verbose:
        print("using %i points on the sky" % ra.size)
        print("using %i mjds" % mjds.size)

    # Set up the sky brightness model
    sm = sb.SkyModel(mags=True, airmass_limit=airmass_limit)

    filter_names = ["u", "g", "r", "i", "z", "y"]

    # Initialize the relevant lists
    dict_of_lists = {
        "mjds": [],
    }
    sky_brightness = {}
    for filter_name in filter_names:
        sky_brightness[filter_name] = []

    length = mjds[-1] - mjds[0]
    last_5_mags = []
    last_5_mjds = []
    for mjd in mjds:
        progress = (mjd - mjd0) / length * 100
        text = "\rprogress = %.1f%%" % progress
        sys.stdout.write(text)
        sys.stdout.flush()
        sm.set_ra_dec_mjd(ra, dec, mjd, degrees=True)
        if sm.sun_alt <= sunLimit_rad:
            mags = sm.return_mags()
            for key in filter_names:
                sky_brightness[key].append(mags[key])
            dict_of_lists["mjds"].append(mjd)
            last_5_mjds.append(mjd)
            last_5_mags.append(mags)
            if len(last_5_mjds) > 5:
                del last_5_mjds[0]
                del last_5_mags[0]

            if np.size(dict_of_lists["mjds"]) > 3:
                if dict_of_lists["mjds"][-2] not in required_mjds:
                    # Check if we can interpolate the second to
                    # last sky brightnesses
                    if dict_of_lists["mjds"][-1] - dict_of_lists["mjds"][-3] < timestep_max:
                        can_interp = True
                        for mjd2 in last_5_mjds:
                            mjd1 = dict_of_lists["mjds"][-3]
                            mjd3 = dict_of_lists["mjds"][-1]
                            if (mjd2 > mjd1) & (mjd2 < mjd3):
                                indx = np.min(np.where(last_5_mjds == mjd2)[0])
                                # Linear interpolation weights
                                wterm = (mjd2 - mjd1) / (mjd3 - mjd1)
                                w1 = 1.0 - wterm
                                w2 = wterm
                                for filter_name in filter_names:
                                    interp_sky = w1 * sky_brightness[filter_name][-3]
                                    interp_sky += w2 * sky_brightness[filter_name][-1]
                                    diff = np.abs(last_5_mags[indx][filter_name] - interp_sky)
                                    if np.size(diff[~np.isnan(diff)]) > 0:
                                        if np.max(diff[~np.isnan(diff)]) > dm:
                                            can_interp = False
                        if can_interp:
                            for key in dict_of_lists:
                                del dict_of_lists[key][-2]
                            for key in sky_brightness:
                                del sky_brightness[key][-2]
    print("")

    final_mjds = np.array(dict_of_lists["mjds"])
    final_sky_mags = np.zeros(
        (final_mjds.size, sky_brightness["r"][0].size),
        dtype=list(zip(filter_names, ["float"] * 6)),
    )
    for key in sky_brightness:
        final_sky_mags[key] = sky_brightness[key]

    import rubin_sim

    version = rubin_sim.version.__version__
    fingerprint = version
    # Generate a header to save all the kwarg info for this run
    if outpath is None:
        outpath = ""
    header = {
        "mjd0": mjd0,
        "mjd_max": mjd_max,
        "timestep": timestep,
        "timestep_max": timestep_max,
        "outfile": outfile,
        "outpath": outpath,
        "nside": nside,
        "verbose": verbose,
        "version": version,
        "fingerprint": fingerprint,
    }

    # Save mjd and sky brightness arrays to an hdf5 file
    hf = h5py.File(outfile, "w")
    hf.create_dataset("mjds", data=final_mjds)
    hf.create_dataset("sky_mags", data=final_sky_mags, compression="gzip")
    hf.create_dataset("timestep_max", data=timestep_max)
    hf.attrs.update(header)
    hf.close()


if __name__ == "__main__":
    # Make a quick small one for speed loading
    print("generating small file")
    m0 = utils.SURVEY_START_MJD
    generate_sky(mjd0=m0 - 1, mjd_max=m0 + 31)

    nyears = 25.0  # 20  # 13
    day_pad = 30

    # Full year
    # mjds = np.arange(59560, 59560+365.25*nyears+day_pad+366, 366)
    # 6-months
    mjds = np.arange(59560, 59560 + 366 * nyears + 366 / 2.0, 366 / 2.0)
    # mjds = [59560, 59563.5]
    # mjds = [60218, 60226]
    # Add some time ahead of time for ComCam
    # nyears = 3
    # mjds = np.arange(58462, 58462+366*nyears+366/2., 366/2.)
    count = 0
    for mjd1, mjd2 in zip(mjds[:-1], mjds[1:]):
        print("Generating file %i" % count)
        # generate_sky(mjd0=mjd1, mjd_max=mjd2+day_pad,
        # outpath='opsimFields', fieldID=True)
        generate_sky(mjd0=mjd1, mjd_max=mjd2 + day_pad)
        count += 1



import unittest

import numpy as np

import rubin_sim.maf.metrics as metrics


class TestSimpleMetrics(unittest.TestCase):
    def test_color_slope(self):
        names = ["night", "observationStartMJD", "filter", "fiveSigmaDepth"]
        types = [int, float, "<U1", float]

        data = np.zeros(4, dtype=list(zip(names, types)))

        # same filter, same night
        data["observationStartMJD"] = np.array([0, 0.25, 0.5, 0.55]) / 24
        data["filter"] = ["r", "r", "r", "r"]
        data["fiveSigmaDepth"] = 25.0

        csm = metrics.ColorSlopeMetric(color_length=1.0, slope_length=3.0)

        cs2n = metrics.ColorSlope2NightMetric(color_length=1.0, slope_length=15.0)
        assert csm.run(data) == 0
        assert cs2n.run(data) == 0

        # diff filter, same night
        # has color, but no slope
        data["observationStartMJD"] = np.array([0, 0.25, 0.5, 0.55]) / 24
        data["filter"] = ["r", "g", "r", "r"]

        assert csm.run(data) == 0
        assert cs2n.run(data) == 0

        # diff filter, same night
        # slope on 1st night, not second
        data["observationStartMJD"] = np.array([0, 0.25, 0.5, 3.55]) / 24
        data["filter"] = ["r", "g", "r", "r"]

        assert csm.run(data) == 1
        assert cs2n.run(data) == 0

        # diff filter, diff night
        # slope on 2nd night, not first
        data["night"] = [0, 0, 0, 1]
        data["observationStartMJD"] = np.array([0, 0.25, 0.5, 25]) / 24
        data["filter"] = ["r", "g", "r", "r"]

        assert csm.run(data) == 0
        assert cs2n.run(data) == 1

        # diff filter, diff night
        # slope on both nights
        data["night"] = [0, 0, 0, 1]
        data["observationStartMJD"] = np.array([0, 0.25, 3.5, 25]) / 24
        data["filter"] = ["r", "g", "r", "r"]

        assert csm.run(data) == 1
        assert cs2n.run(data) == 1

        # diff filter, diff night
        # slope on both nights, but no color
        data["night"] = [0, 0, 0, 1]
        data["observationStartMJD"] = np.array([0, 5.25, 3.5, 25]) / 24
        data["filter"] = ["r", "g", "r", "r"]

        assert csm.run(data) == 0
        assert cs2n.run(data) == 0


if __name__ == "__main__":
    unittest.main()



__all__ = ("MetricVsH", "MetricVsOrbit", "MetricVsOrbitPoints")


import matplotlib.cm as cm
import matplotlib.pyplot as plt
import numpy as np

from .plot_handler import BasePlotter

# mag_sun = -27.1
# apparent r band magnitude of the sun.
# this sets the band for the magnitude limit.
# see http://www.ucolick.org/~cnaw/sun.html for apparent mags in other bands.
mag_sun = -26.74
# apparent V band magnitude of the Sun (our H mags translate to V band)
km_per_au = 1.496e8
m_per_km = 1000


class MetricVsH(BasePlotter):
    """
    Plot metric values versus H.
    Marginalize over metric values in each H bin using 'np_reduce'.
    """

    def __init__(self):
        self.plot_type = "MetricVsH"
        self.object_plotter = False
        self.default_plot_dict = {
            "title": None,
            "xlabel": "H (mag)",
            "ylabel": None,
            "label": None,
            "np_reduce": None,
            "nbins": None,
            "albedo": None,
            "Hmark": None,
            "HmarkLinestyle": ":",
            "figsize": None,
        }
        self.min_hrange = 1.0

    def __call__(self, metric_value, slicer, user_plot_dict, fig=None):
        if "linestyle" not in user_plot_dict:
            user_plot_dict["linestyle"] = "-"
        plot_dict = {}
        plot_dict.update(self.default_plot_dict)
        plot_dict.update(user_plot_dict)
        hvals = slicer.slice_points["H"]
        reduce_func = plot_dict["np_reduce"]
        if reduce_func is None:
            reduce_func = np.mean
        if hvals.shape[0] == 1:
            # We have a simple set of values to plot against H.
            # This may be due to running a summary metric,
            # such as completeness.
            m_vals = metric_value[0].filled()
        elif len(hvals) == slicer.shape[1]:
            # Using cloned H distribution.
            # Apply 'np_reduce' method directly to metric values,
            # and plot at matching H values.
            m_vals = reduce_func(metric_value.filled(), axis=0)
        else:
            # Probably each object has its own H value.
            hrange = hvals.max() - hvals.min()
            min_h = hvals.min()
            if hrange < self.min_hrange:
                hrange = self.min_hrange
                min_h = hvals.min() - hrange / 2.0
            nbins = plot_dict["nbins"]
            if nbins is None:
                nbins = 30
            stepsize = hrange / float(nbins)
            bins = np.arange(min_h, min_h + hrange + stepsize / 2.0, stepsize)
            # In each bin of H, calculate the 'np_reduce' value of the
            # corresponding metric_values.
            inds = np.digitize(hvals, bins)
            inds = inds - 1
            m_vals = np.zeros(len(bins), float)
            for i in range(len(bins)):
                match = metric_value[inds == i]
                if len(match) == 0:
                    m_vals[i] = slicer.badval
                else:
                    m_vals[i] = reduce_func(match.filled())
            hvals = bins
        # Plot the values.
        if fig is None:
            fig = plt.figure(figsize=plot_dict["figsize"])
        ax = plt.gca()
        ax.plot(
            hvals,
            m_vals,
            color=plot_dict["color"],
            linestyle=plot_dict["linestyle"],
            label=plot_dict["label"],
        )
        if "x_min" in plot_dict:
            ax.set_xlim(left=plot_dict["x_min"])
        if "x_max" in plot_dict:
            ax.set_xlim(right=plot_dict["x_max"])
        if "y_min" in plot_dict:
            ax.set_ylim(bottom=plot_dict["y_min"])
        if "y_max" in plot_dict:
            ax.set_ylim(top=plot_dict["y_max"])
        # Convert hvals to diameter, using 'albedo' - add these upper xticks
        albedo = plot_dict["albedo"]
        y = 1.0
        if albedo is not None:
            ax2 = ax.twiny()
            hmin, hmax = ax.get_xlim()
            dmax = 2.0 * np.sqrt(10 ** ((mag_sun - hmin - 2.5 * np.log10(albedo)) / 2.5))
            dmin = 2.0 * np.sqrt(10 ** ((mag_sun - hmax - 2.5 * np.log10(albedo)) / 2.5))
            dmax = dmax * km_per_au * m_per_km
            dmin = dmin * km_per_au * m_per_km
            ax2.set_xlim(dmax, dmin)
            ax2.set_xscale("log")
            dmid = (dmax - dmin) / 2 + dmin
            dmid = np.power(10, round(np.log10(dmid)))
            dmin = np.power(10, round(np.log10(dmin)))
            dmax = np.power(10, round(np.log10(dmax)))
            dticks = np.array([dmin, dmid, dmax])
            ax2.set_xticks(dticks)
            ax2.set_xlabel("D (m)", labelpad=-5, horizontalalignment="right")
            ax2.grid(False)
            plt.sca(ax)
            y = 1.1
        plt.grid(True)
        if plot_dict["Hmark"] is not None:
            plt.axvline(
                x=plot_dict["Hmark"],
                color="r",
                linestyle=plot_dict["HmarkLinestyle"],
                alpha=0.3,
            )
        plt.title(plot_dict["title"], y=y)
        plt.xlabel(plot_dict["xlabel"])
        plt.ylabel(plot_dict["ylabel"])
        plt.tight_layout()
        return fig


class MetricVsOrbit(BasePlotter):
    """
    Plot metric values (at a particular H value) vs. orbital parameters.
    Marginalize over metric values in each orbital bin using 'np_reduce'.
    """

    def __init__(self, xaxis="q", yaxis="e"):
        self.plot_type = "MetricVsOrbit_%s%s" % (xaxis, yaxis)
        self.object_plotter = False
        self.default_plot_dict = {
            "title": None,
            "xlabel": xaxis,
            "ylabel": yaxis,
            "xaxis": xaxis,
            "yaxis": yaxis,
            "label": None,
            "cmap": cm.viridis,
            "np_reduce": None,
            "nxbins": None,
            "nybins": None,
            "levels": None,
            "h_val": None,
            "Hwidth": None,
            "figsize": None,
        }

    def __call__(self, metric_value, slicer, user_plot_dict, fig=None):
        plot_dict = {}
        plot_dict.update(self.default_plot_dict)
        plot_dict.update(user_plot_dict)
        if fig is None:
            fig = plt.figure(figsize=plot_dict["figsize"])
        xvals = slicer.slice_points["orbits"][plot_dict["xaxis"]]
        yvals = slicer.slice_points["orbits"][plot_dict["yaxis"]]
        # Set x/y bins.
        nxbins = plot_dict["nxbins"]
        nybins = plot_dict["nybins"]
        if nxbins is None:
            nxbins = 100
        if nybins is None:
            nybins = 100
        if "xbins" in plot_dict:
            xbins = plot_dict["xbins"]
        else:
            xbinsize = (xvals.max() - xvals.min()) / float(nxbins)
            xbins = np.arange(xvals.min(), xvals.max() + xbinsize / 2.0, xbinsize)
        if "ybins" in plot_dict:
            ybins = plot_dict["ybins"]
        else:
            ybinsize = (yvals.max() - yvals.min()) / float(nybins)
            ybins = np.arange(yvals.min(), yvals.max() + ybinsize / 2.0, ybinsize)
        nxbins = len(xbins)
        nybins = len(ybins)
        # Identify the relevant metric_values for the Hvalue we want to plot.
        hvals = slicer.slice_points["H"]
        hwidth = plot_dict["hwidth"]
        if hwidth is None:
            hwidth = 1.0
        if len(hvals) == slicer.shape[1]:
            if plot_dict["hval"] is None:
                hidx = int(len(hvals) / 2)
                hval = hvals[hidx]
            else:
                hval = plot_dict["hval"]
                hidx = np.where(np.abs(hvals - hval) == np.abs(hvals - hval).min())[0]
                hidx = hidx[0]
        else:
            if plot_dict["hval"] is None:
                hval = np.median(hvals)
                hidx = np.where(np.abs(hvals - hval) <= hwidth / 2.0)[0]
            else:
                hval = plot_dict["hvals"]
                hidx = np.where(np.abs(hvals - hval) <= hwidth / 2.0)[0]
        if len(hvals) == slicer.shape[1]:
            m_vals = np.swapaxes(metric_value, 1, 0)[hidx].filled()
        else:
            m_vals = metric_value[hidx].filled()
        # Calculate the np_reduce'd metric values at each x/y bin.
        if "color_min" in plot_dict:
            badval = plot_dict["color_min"] - 1
        else:
            badval = slicer.badval
        binvals = np.zeros((nybins, nxbins), dtype="float") + badval
        xidxs = np.digitize(xvals, xbins) - 1
        yidxs = np.digitize(yvals, ybins) - 1
        reduce_func = plot_dict["np_reduce"]
        if reduce_func is None:
            reduce_func = np.mean
        for iy in range(nybins):
            ymatch = np.where(yidxs == iy)[0]
            for ix in range(nxbins):
                xmatch = np.where(xidxs[ymatch] == ix)[0]
                match_vals = m_vals[ymatch][xmatch]
                if len(match_vals) > 0:
                    binvals[iy][ix] = reduce_func(match_vals)
        xi, yi = np.meshgrid(xbins, ybins)
        if "color_min" in plot_dict:
            v_min = plot_dict["color_min"]
        else:
            v_min = binvals.min()
        if "color_max" in plot_dict:
            v_max = plot_dict["color_max"]
        else:
            v_max = binvals.max()
        nlevels = plot_dict["levels"]
        if nlevels is None:
            nlevels = 200
        levels = np.arange(v_min, v_max, (v_max - v_min) / float(nlevels))
        plt.contourf(xi, yi, binvals, levels, extend="max", zorder=0, cmap=plot_dict["cmap"])
        cbar = plt.colorbar()
        label = plot_dict["label"]
        if label is None:
            label = ""
        cbar.set_label(label + " @ H=%.1f" % (hval))
        plt.title(plot_dict["title"])
        plt.xlabel(plot_dict["xlabel"])
        plt.ylabel(plot_dict["ylabel"])
        return fig


class MetricVsOrbitPoints(BasePlotter):
    """
    Plot metric values (at a particular H value) as function
    of orbital parameters, using points for each metric value.
    """

    def __init__(self, xaxis="q", yaxis="e"):
        self.plot_type = "MetricVsOrbit"
        self.object_plotter = False
        self.default_plot_dict = {
            "title": None,
            "xlabel": xaxis,
            "ylabel": yaxis,
            "label": None,
            "cmap": cm.viridis,
            "xaxis": xaxis,
            "yaxis": yaxis,
            "h_val": None,
            "Hwidth": None,
            "foregroundPoints": True,
            "backgroundPoints": False,
            "figsize": None,
        }

    def __call__(self, metric_value, slicer, user_plot_dict, fig=None):
        plot_dict = {}
        plot_dict.update(self.default_plot_dict)
        plot_dict.update(user_plot_dict)
        if fig is None:
            fig = plt.figure(figsize=plot_dict["figsize"])
        xvals = slicer.slice_points["orbits"][plot_dict["xaxis"]]
        yvals = slicer.slice_points["orbits"][plot_dict["yaxis"]]
        # Identify the relevant metric_values for the Hvalue we want to plot.
        hvals = slicer.slice_points["H"]
        hwidth = plot_dict["hwidth"]
        if hwidth is None:
            hwidth = 1.0
        if len(hvals) == slicer.shape[1]:
            if plot_dict["hval"] is None:
                hidx = int(len(hvals) / 2)
                hval = hvals[hidx]
            else:
                hval = plot_dict["hval"]
                hidx = np.where(np.abs(hvals - hval) == np.abs(hvals - hval).min())[0]
                hidx = hidx[0]
        else:
            if plot_dict["hval"] is None:
                hval = np.median(hvals)
                hidx = np.where(np.abs(hvals - hval) <= hwidth / 2.0)[0]
            else:
                hval = plot_dict["hvals"]
                hidx = np.where(np.abs(hvals - hval) <= hwidth / 2.0)[0]
        if len(hvals) == slicer.shape[1]:
            m_vals = np.swapaxes(metric_value, 1, 0)[hidx]
        else:
            m_vals = metric_value[hidx]
        if "color_min" in plot_dict:
            v_min = plot_dict["color_min"]
        else:
            v_min = m_vals.min()
        if "color_max" in plot_dict:
            v_max = plot_dict["color_max"]
        else:
            v_max = m_vals.max()
        if plot_dict["backgroundPoints"]:
            # This isn't quite right for the condition .. but will do for now.
            condition = np.where(m_vals == 0)
            plt.plot(
                xvals[condition],
                yvals[condition],
                "r.",
                markersize=4,
                alpha=0.5,
                zorder=3,
            )
        if plot_dict["foregroundPoints"]:
            plt.scatter(
                xvals,
                yvals,
                c=m_vals,
                vmin=v_min,
                vmax=v_max,
                cmap=plot_dict["cmap"],
                s=15,
                alpha=0.8,
                zorder=0,
            )
            cbar = plt.colorbar()
            label = plot_dict["label"]
            if label is None:
                label = ""
        cbar.set_label(label + " @ H=%.1f" % (hval))
        plt.title(plot_dict["title"])
        plt.xlabel(plot_dict["xlabel"])
        plt.ylabel(plot_dict["ylabel"])
        return fig



__all__ = ("XyPlotter",)

import matplotlib.pyplot as plt

from .plot_handler import BasePlotter


class XyPlotter(BasePlotter):
    """Bare-bones plotter for making scatter plots. Expects single metric value
    (e.g, from UniSlicer or UserPointSlicer with one point)"""

    def __init__(self):
        self.object_plotter = True
        self.plot_type = "simple"
        self.default_plot_dict = {
            "title": None,
            "xlabel": "",
            "ylabel": "",
            "figsize": None,
        }

    def __call__(self, metric_value_in, slicer, user_plot_dict, fig=None):
        plot_dict = {}
        plot_dict.update(self.default_plot_dict)
        plot_dict.update(user_plot_dict)
        plot_dict.update(metric_value_in[0]["plot_dict"])

        if fig is None:
            fig = plt.figure(figsize=plot_dict["figsize"])
        ax = fig.add_subplot(111)
        x = metric_value_in[0]["x"]
        y = metric_value_in[0]["y"]
        ax.plot(x, y)
        ax.set_title(plot_dict["title"])
        ax.set_xlabel(plot_dict["xlabel"])
        ax.set_ylabel(plot_dict["ylabel"])
        return fig



__all__ = ("HourglassSlicer",)


from rubin_sim.maf.plots import HourglassPlot

from .uni_slicer import UniSlicer


class HourglassSlicer(UniSlicer):
    """Slicer to make the filter hourglass plots"""

    def __init__(self, verbose=True, badval=-666):
        # Inherits from UniSlicer, so nslice=1 and only one 'slice'.
        super(HourglassSlicer, self).__init__(verbose=verbose, badval=badval)
        self.columns_needed = []
        self.slicer_name = "HourglassSlicer"
        self.plot_funcs = [
            HourglassPlot,
        ]

    def write_data(self, outfilename, metric_values, metric_name="", **kwargs):
        """
        Override base write method: we don't want to save hourglass metric
        data.

        The data volume is too large.
        """
        pass

    def read_metric_data(self, infilename):
        """
        Override base read method to 'pass':
        we don't save or read hourglass metric data.

        The data volume is too large.
        """
        pass



import os
import shutil
import tempfile
import unittest

import numpy as np
from rubin_scheduler.utils.code_utilities import sims_clean_up

import rubin_sim.maf.batches as batches
import rubin_sim.maf.db as db
import rubin_sim.maf.metric_bundles as metric_bundles
from rubin_sim.data import get_data_dir
from rubin_sim.maf.slicers import MoObjSlicer

TEST_DB = "example_v3.4_0yrs.db"


class TestBatches(unittest.TestCase):
    @classmethod
    def tearDown_class(cls):
        sims_clean_up()

    def setUp(self):
        self.out_dir = tempfile.mkdtemp(prefix="TMB")

    @unittest.skipUnless(
        os.path.isdir(os.path.join(get_data_dir(), "maf")),
        "Skip these batches unless MAF data present, required for setup",
    )
    def testload_them_all(self):
        ack = batches.altazHealpix()
        assert ack is not None
        ack = batches.altazLambert()
        assert ack is not None
        ack = batches.astrometryBatch()
        assert ack is not None
        ack = batches.ddfBatch()
        assert ack is not None
        ack = batches.fOBatch()
        assert ack is not None
        ack = batches.standard_summary()
        assert ack is not None
        ack = batches.standard_metrics("night")
        assert ack is not None
        ack = batches.filtersPerNight()
        assert ack is not None
        ack = batches.filtersWholeSurvey()
        assert ack is not None
        ack = batches.glanceBatch()
        assert ack is not None
        ack = batches.hourglassPlots()
        assert ack is not None
        ack = batches.interNight()
        assert ack is not None
        ack = batches.intraNight()
        assert ack is not None
        ack = batches.metadataBasics("airmass")
        assert ack is not None
        ack = batches.metadataBasicsAngle("rotskyPos")
        assert ack is not None
        ack = batches.metadataMaps("fiveSigmaDepth")
        assert ack is not None
        ack = batches.nvisitsM5Maps()
        assert ack is not None
        ack = batches.nvisitsPerNight()
        assert ack is not None
        ack = batches.nvisitsPerSubset()
        assert ack is not None
        ack = batches.openshutterFractions()
        assert ack is not None
        ack = batches.rapidRevisitBatch()
        assert ack is not None
        ack = batches.slewBasics()
        assert ack is not None
        ack = batches.tEffMetrics()
        assert ack is not None
        ack = batches.timeGaps()
        assert ack is not None

    @unittest.skipUnless(
        os.path.isdir(os.path.join(get_data_dir(), "maf")),
        "Skip these batches unless MAF data present, required for setup",
    )
    def test_moving_objects_batches(self):
        slicer = MoObjSlicer()
        ack = batches.quick_discovery_batch(slicer)
        assert ack is not None
        ack = batches.discovery_batch(slicer)
        assert ack is not None
        ack = batches.characterization_inner_batch(slicer)
        assert ack is not None
        ack = batches.characterization_outer_batch(slicer)
        assert ack is not None

    def test_moving_fractions(self):
        bdict = {}
        bundle = metric_bundles.create_empty_mo_metric_bundle()
        bundle.run_name = "dummy_name"
        bdict["days"] = bundle
        bdict["days"].slicer = MoObjSlicer()
        bdict["days"].slicer.slice_points["H"] = np.arange(50)
        # This is a start, but the methods below aren't really doing anything
        # without appropriately named/values bundles in the bdict.
        batches.run_completeness_summary(bdict, 0.0, [10, 20], None, None)
        batches.run_fraction_summary(bdict, 0.0, None, None)
        batches.plot_fractions(bdict, out_dir=self.out_dir)
        # This batch takes a single bundle because it plots that single
        # bundle with multiple summary interpretations of the metric
        batches.plot_single(bundle, out_dir=self.out_dir)
        batches.plot_activity(bdict)

    @unittest.skipUnless(
        os.path.isdir(os.path.join(get_data_dir(), "maps")),
        "Skipping scienceRadarBatch test because operating without full MAF test data",
    )
    def test_science_radar(self):
        # Loading the science radar batch requires reading a significant
        # set of input files
        # This test is skipped if running with the lighter set of test data.
        # batch requires reading a lot of input files for lightcurves
        ack = batches.science_radar_batch()
        assert ack is not None

    @unittest.skipUnless(
        os.path.isdir(os.path.join(get_data_dir(), "maf")),
        "Skipping glance test because operating without full MAF test data",
    )
    def test_glance(self):
        ack = batches.glanceBatch()
        database = os.path.join(get_data_dir(), "tests", TEST_DB)
        results_db = db.ResultsDb(out_dir=self.out_dir)
        bgroup = metric_bundles.MetricBundleGroup(ack, database, out_dir=self.out_dir, results_db=results_db)
        bgroup.run_all()

    def tearDown(self):
        if os.path.isdir(self.out_dir):
            shutil.rmtree(self.out_dir)


if __name__ == "__main__":
    unittest.main()



import os
import unittest

import numpy as np
from rubin_scheduler.data import get_data_dir
from rubin_scheduler.utils.code_utilities import sims_clean_up

from rubin_sim.phot_utils.bandpass import Bandpass
from rubin_sim.phot_utils.sed import Sed


class PhotometryUnitTest(unittest.TestCase):
    @classmethod
    def tearDown_class(cls):
        sims_clean_up()

    def test_alternate_bandpasses_stars(self):
        """Test our ability to do photometry using non-LSST bandpasses.

        Calculate the photometry by built-in methods and 'by hand'.
        """
        bandpass_dir = os.path.join(get_data_dir(), "tests", "cartoonSedTestData")

        test_band_passes = {}
        keys = ["u", "g", "r", "i", "z"]

        bplist = []

        for kk in keys:
            test_band_passes[kk] = Bandpass()
            test_band_passes[kk].read_throughput(os.path.join(bandpass_dir, "test_bandpass_%s.dat" % kk))
            bplist.append(test_band_passes[kk])

        sed_obj = Sed()
        phi_array, wave_len_step = sed_obj.setup_phi_array(bplist)

        sed_file_name = os.path.join(get_data_dir(), "tests", "cartoonSedTestData/starSed/")
        sed_file_name = os.path.join(sed_file_name, "kurucz", "km20_5750.fits_g40_5790.gz")
        ss = Sed()
        ss.read_sed_flambda(sed_file_name)

        control_bandpass = Bandpass()
        control_bandpass.imsim_bandpass()
        ff = ss.calc_flux_norm(22.0, control_bandpass)
        ss.multiply_flux_norm(ff)

        test_mags = []
        for kk in keys:
            test_mags.append(ss.calc_mag(test_band_passes[kk]))

        ss.resample_sed(wavelen_match=bplist[0].wavelen)
        ss.flambda_tofnu()
        mags = -2.5 * np.log10(np.sum(phi_array * ss.fnu, axis=1) * wave_len_step) - ss.zp
        self.assertEqual(len(mags), len(test_mags))
        self.assertGreater(len(mags), 0)
        for j in range(len(mags)):
            self.assertAlmostEqual(mags[j], test_mags[j], 3)


if __name__ == "__main__":
    unittest.main()



import glob
import os
import shutil
import tempfile
import unittest

from rubin_scheduler.data import get_data_dir
from rubin_scheduler.utils.code_utilities import sims_clean_up

import rubin_sim.maf.db as db
import rubin_sim.maf.maps as maps
import rubin_sim.maf.metric_bundles as metric_bundles
import rubin_sim.maf.metrics as metrics
import rubin_sim.maf.slicers as slicers
import rubin_sim.maf.stackers as stackers

TEST_DB = "example_v3.4_0yrs.db"


class TestMetricBundle(unittest.TestCase):
    @classmethod
    def tearDown_class(cls):
        sims_clean_up()

    def setUp(self):
        self.out_dir = tempfile.mkdtemp(prefix="TMB")

    def test_out(self):
        """
        Check that the metric bundle can generate the expected output
        """
        nside = 8
        slicer = slicers.HealpixSlicer(nside=nside)
        metric = metrics.MeanMetric(col="airmass")
        sql = 'filter="r"'
        stacker1 = stackers.HourAngleStacker()
        stacker2 = stackers.GalacticStacker()
        map = maps.GalCoordsMap()

        metric_b = metric_bundles.MetricBundle(
            metric, slicer, sql, stacker_list=[stacker1, stacker2], maps_list=[map]
        )
        database = os.path.join(get_data_dir(), "tests", TEST_DB)

        results_db = db.ResultsDb(out_dir=self.out_dir)

        bgroup = metric_bundles.MetricBundleGroup(
            {0: metric_b}, database, out_dir=self.out_dir, results_db=results_db
        )
        bgroup.run_all()
        bgroup.plot_all()
        bgroup.write_all()

        out_thumbs = glob.glob(os.path.join(self.out_dir, "thumb*"))
        out_npz = glob.glob(os.path.join(self.out_dir, "*.npz"))
        out_pdf = glob.glob(os.path.join(self.out_dir, "*.pdf"))

        # By default, make 2 plots for healpix
        assert len(out_thumbs) == 2
        assert len(out_pdf) == 2
        assert len(out_npz) == 1

    def tearDown(self):
        if os.path.isdir(self.out_dir):
            shutil.rmtree(self.out_dir)


if __name__ == "__main__":
    unittest.main()



__all__ = ("eb_vhp",)

import os

import healpy as hp
import numpy as np
from rubin_scheduler.data import get_data_dir

from rubin_sim.maf.utils import radec2pix


def eb_vhp(nside, ra=None, dec=None, pixels=None, interp=False, map_path=None):
    """Read in a healpix dust map and return values for given RA, Dec values.

    This is primarily a tool for the rubin_sim.maf.DustMap class.

    nside : `int`
        Healpixel resolution (2^x).
    ra : `np.ndarray` or `float`, opt
        RA (can take numpy array).
        Default None sets up healpix array of nside. Radians.
    dec : `np.ndarray` or `float`, opt
        Dec (can take numpy array).
        Default None set up healpix array of nside. Radians.
    pixels : `np.ndarray`, opt
        Healpixel IDs, to sub-select particular healpix points.
        Default uses all points.
        NOTE - to use a healpix map, set pixels and not ra/dec.
    interp : `bool`, opt
        Should returned values be interpolated (True)
        or just nearest neighbor (False)
    map_path : `str`, opt
        Path to directory containing dust map files.
    """

    if (ra is None) & (dec is None) & (pixels is None):
        raise RuntimeError("Need to set ra,dec or pixels.")

    # Load the map
    if map_path is not None:
        ebv_data_dir = map_path
    else:
        ebv_data_dir = os.path.join(get_data_dir(), "maps", "DustMaps")
    if not hasattr(eb_vhp, "nside"):
        eb_vhp.nside = nside

    if (not hasattr(eb_vhp, "dustmap")) | (eb_vhp.nside != nside):
        eb_vhp.nside = nside
        filename = "dust_nside_%i.npz" % eb_vhp.nside
        eb_vhp.dustMap = np.load(os.path.join(ebv_data_dir, filename))["ebvMap"]

    # If we are interpolating to arbitrary positions
    if interp:
        result = hp.get_interp_val(eb_vhp.dustMap, np.pi / 2.0 - dec, ra)
    else:
        # If we know the pixel indices we want
        if pixels is not None:
            result = eb_vhp.dustMap[pixels]
        # Look up
        else:
            pixels = radec2pix(eb_vhp.nside, ra, dec)
            result = eb_vhp.dustMap[pixels]

    return result



from .base_obs import *  # noqa: F403
from .cheby_fits import *  # noqa: F403
from .cheby_values import *  # noqa: F403
from .chebyshev_utils import *  # noqa: F403
from .direct_obs import *  # noqa: F403
from .ooephemerides import *  # noqa: F403
from .orbits import *  # noqa: F403
from .utils import *  # noqa: F403



# Variability Depth Metric
# Keaton Bell (keatonb@astro.as.utexas.edu)

__all__ = ("VarDepth",)

import numpy as np
from scipy.interpolate import UnivariateSpline
from scipy.stats import chi2

from rubin_sim.maf.metrics import BaseMetric


class VarDepth(BaseMetric):
    """Calculate the survey depth that a variable star can be
    reliably identified.

    Parameters
    ----------
    completeness : `float`, opt
        Fractional desired completeness of recovered variable sample.
    contamination : `float`, opt
        Fractional allowed incompleteness of recovered nonvariables.
    numruns : `int`, opt
        Number of simulated realizations of noise.
        Most computationally expensive part of metric.
    signal : `float`, opt
        Sqrt total pulsational power meant to be recovered.
    magres : `float`, opt
        desired resolution of variability depth result.
    """

    def __init__(
        self,
        m5_col="fiveSigmaDepth",
        metric_name="variability depth",
        completeness=0.95,
        contamination=0.05,
        numruns=10000,
        signal=0.01,
        magres=0.01,
        **kwargs,
    ):
        self.m5col = m5_col
        self.completeness = completeness
        self.contamination = contamination
        self.numruns = numruns
        self.signal = signal
        self.magres = magres
        super(VarDepth, self).__init__(col=m5_col, metric_name=metric_name, **kwargs)

    def run(self, data_slice, slice_point=None):
        # Get the visit information
        m5 = data_slice[self.m5col]
        # Number of visits
        N = len(m5)

        # magnitudes to be sampled
        mag = np.arange(16, np.mean(m5), 0.5)
        # hold the distance between the completeness and contamination goals.
        res = np.zeros(mag.shape)
        # make them nans for now
        res[:] = np.nan

        # hold the measured noise-only variances
        noiseonlyvar = np.zeros(self.numruns)

        # Calculate the variance at a reference magnitude and scale from that
        m0 = 20.0
        sigmaref = 0.2 * (10.0 ** (-0.2 * m5)) * (10.0 ** (0.2 * m0))

        # run the simulations
        # Simulate the measured noise-only variances at a reference magnitude
        for i in np.arange(self.numruns):
            # random realization of the Gaussian error distributions
            scatter = np.random.randn(N) * sigmaref
            noiseonlyvar[i] = np.var(scatter)  # store the noise-only variance

        # Since we are treating the underlying signal being representable by a
        # fixed-width gaussian, its variance pdf is a Chi-squared distribution
        # with the degrees of freedom=visits. Since variances add, the variance
        # pdfs convolve. The cumulative distribution function of the sum of two
        # random deviates is the convolution of one pdf with a cdf.

        # We'll consider the cdf of the noise-only variances
        # because it's easier to interpolate
        noisesorted = np.sort(noiseonlyvar)
        # linear interpolation
        interpnoisecdf = UnivariateSpline(
            noisesorted, np.arange(self.numruns) / float(self.numruns), k=1, s=0
        )

        # We need a binned, signal-only variance probability
        # distribution function for numerical convolution
        numsignalsamples = 100
        xsig = np.linspace(chi2.ppf(0.001, N), chi2.ppf(0.999, N), numsignalsamples)
        signalpdf = chi2.pdf(xsig, N)
        # correct x to the proper variance scale
        xsig = (self.signal**2.0) * xsig / N
        pdfstepsize = xsig[1] - xsig[0]
        # Since everything is going to use this stepsize down the line,
        # normalize so the pdf integrates to 1 when summed
        # (no factor of stepsize needed)
        signalpdf /= np.sum(signalpdf)

        # run through the sample magnitudes, calculate distance between cont
        # and comp thresholds.
        # run until solution found.
        solutionfound = False

        for i, mref in enumerate(mag):
            # i counts and mref is the currently sampled magnitude
            # Scale factor from m0
            scalefact = 10.0 ** (0.4 * (mref - m0))

            # Calculate the desired contamination threshold
            contthresh = np.percentile(noiseonlyvar, 100.0 - 100.0 * self.contamination) * scalefact

            # Realize the noise CDF at the required stepsize
            xnoise = np.arange(noisesorted[0] * scalefact, noisesorted[-1] * scalefact, pdfstepsize)

            # Only do calculation if near the solution:
            if (len(xnoise) > numsignalsamples / 10) and (not solutionfound):
                noisecdf = interpnoisecdf(xnoise / scalefact)
                # turn into a noise pdf
                noisepdf = noisecdf[1:] - noisecdf[:-1]
                noisepdf /= np.sum(noisepdf)
                # from cdf to pdf conversion
                xnoise = (xnoise[1:] + xnoise[:-1]) / 2.0

                # calculate and plot the convolution =
                # signal+noise variance dist.
                convolution = 0
                if len(noisepdf) > len(signalpdf):
                    convolution = np.convolve(noisepdf, signalpdf)
                else:
                    convolution = np.convolve(signalpdf, noisepdf)
                xconvolved = xsig[0] + xnoise[0] + np.arange(len(convolution)) * pdfstepsize

                # calculate the completeness threshold
                combinedcdf = np.cumsum(convolution)
                findcompthresh = UnivariateSpline(combinedcdf, xconvolved, k=1, s=0)
                compthresh = findcompthresh(1.0 - self.completeness)

                res[i] = compthresh - contthresh
                if res[i] < 0:
                    solutionfound = True

        # interpolate for where the thresholds coincide
        # print res
        if np.sum(np.isfinite(res)) > 1:
            f1 = UnivariateSpline(mag[np.isfinite(res)], res[np.isfinite(res)], k=1, s=0)
            # sample the magnitude range at given resolution
            magsamples = np.arange(16, np.mean(m5), self.magres)
            vardepth = magsamples[np.argmin(np.abs(f1(magsamples)))]
            return vardepth
        else:
            return min(mag) - 1



import unittest

import rubin_sim.maf.metrics as metrics


class TestBaseMetric(unittest.TestCase):
    def test_reduce_dict(self):
        """Test that reduce dictionary is created."""
        testmetric = metrics.BaseMetric("testcol")
        self.assertEqual(list(testmetric.reduce_funcs.keys()), [])

    def test_metric_name(self):
        """Test that metric name is set appropriately automatically
        and when explicitly passed.
        """
        # Test automatic setting of metric name
        testmetric = metrics.BaseMetric("testcol")
        self.assertEqual(testmetric.name, "Base testcol")
        testmetric = metrics.BaseMetric(["testcol1", "testcol2"])
        self.assertEqual(testmetric.name, "Base testcol1, testcol2")
        # Test explicit setting of metric name
        testmetric = metrics.BaseMetric("testcol", metric_name="Test")
        self.assertEqual(testmetric.name, "Test")

    def test_col_registry(self):
        """Test column registry adds to colRegistry as expected"""
        # Clear the registry to make sure we start clear
        colreg = metrics.ColRegistry()
        colreg.clear_reg()

        cols = "onecolumn"
        colset = set()
        colset.add(cols)
        testmetric = metrics.BaseMetric(cols)
        # Class registry should have dictionary with values =
        # set of columns for metric class
        for item in colset:
            self.assertIn(item, testmetric.col_registry.col_set)
        cols = ["onecolumn", "twocolumn"]
        colset.add("twocolumn")
        testmetric = metrics.BaseMetric(cols)
        for item in colset:
            self.assertIn(item, testmetric.col_registry.col_set)
        # Test with additional (different) metric
        cols = "twocolumn"
        testmetric2 = metrics.MeanMetric(cols)
        for item in colset:
            self.assertIn(item, testmetric2.col_registry.col_set)

        # test that the registry can be cleared
        colreg.clear_reg()
        assert len(colreg.col_set) == 0
        assert len(colreg.db_set) == 0
        assert len(colreg.stacker_dict) == 0

    def test_metric_dtype(self):
        """Test that base metric data value type set appropriately"""
        cols = "onecolumn"
        testmetric = metrics.BaseMetric(cols)
        self.assertEqual(testmetric.metric_dtype, "float")
        testmetric = metrics.BaseMetric(cols, metric_dtype="object")
        self.assertEqual(testmetric.metric_dtype, "object")

    def test_units(self):
        """Test unit setting (including units set by utils.getColInfo)"""
        cols = "onecolumn"
        # Test for column not in colInfo, units not set by hand.
        testmetric = metrics.BaseMetric(cols)
        self.assertEqual(testmetric.units, "")
        # Test for column not in colInfo, units set by hand.
        testmetric = metrics.BaseMetric(cols, units="Test")
        self.assertEqual(testmetric.units, "Test")
        # Test for column in colInfo (looking up units in colInfo)
        cols = "finSeeing"
        testmetric = metrics.BaseMetric(cols)
        self.assertEqual(testmetric.units, "arcsec")
        # Test for column in colInfo but units overriden
        testmetric = metrics.BaseMetric(cols, units="Test")
        self.assertEqual(testmetric.units, "Test")
        # Test for multiple columns not in colInfo
        cols = ["onecol", "twocol"]
        testmetric = metrics.BaseMetric(cols)
        self.assertEqual(testmetric.units, "")
        # Test for multiple columns in colInfo
        cols = ["finSeeing", "filtSkyBrightness"]
        testmetric = metrics.BaseMetric(cols)
        self.assertEqual(testmetric.units, "arcsec mag/sq arcsec")
        # Test for multiple columns, only one in colInfo
        cols = ["finSeeing", "twocol"]
        testmetric = metrics.BaseMetric(cols)
        self.assertEqual(testmetric.units, "arcsec ")
        # Test for multiple columns both in colInfo but repeated
        cols = ["finSeeing", "finSeeing"]
        testmetric = metrics.BaseMetric(cols)
        self.assertEqual(testmetric.units, "arcsec arcsec")


if __name__ == "__main__":
    unittest.main()



__all__ = (
    "ss_population_defaults",
    "quick_discovery_batch",
    "discovery_batch",
    "run_completeness_summary",
    "plot_completeness",
    "characterization_inner_batch",
    "characterization_outer_batch",
    "run_fraction_summary",
    "plot_fractions",
    "plot_single",
    "plot_activity",
    "read_and_combine",
    "combine_subsets",
)

import os
from copy import deepcopy

import matplotlib.pyplot as plt
import numpy as np
import numpy.ma as ma

import rubin_sim.maf.metric_bundles as mb
import rubin_sim.maf.metrics as metrics
import rubin_sim.maf.plots as plots
import rubin_sim.maf.stackers as stackers
from rubin_sim.maf.metric_bundles import MoMetricBundle

from .col_map_dict import col_map_dict
from .common import (
    fraction_population_at_threshold,
    summary_completeness_at_time,
    summary_completeness_over_h,
)


def ss_population_defaults(objtype):
    """Provide useful default ranges for H,
    based on objtype of population type.
    """
    defaults = {}
    defaults["Vatira"] = {
        "h_range": [16, 28, 0.2],
        "h_mark": 22,
        "magtype": "asteroid",
        "char": "inner",
    }
    defaults["PHA"] = {
        "h_range": [16, 28, 0.2],
        "h_mark": 22,
        "magtype": "asteroid",
        "char": "inner",
    }
    defaults["NEO"] = {
        "h_range": [16, 28, 0.2],
        "h_mark": 22,
        "magtype": "asteroid",
        "char": "inner",
    }
    defaults["MBA"] = {
        "h_range": [16, 26, 0.2],
        "h_mark": 20,
        "magtype": "asteroid",
        "char": "inner",
    }
    defaults["MBC"] = {
        "h_range": [16, 26, 0.2],
        "h_mark": 20,
        "magtype": "asteroid",
        "char": "inner",
    }
    defaults["Trojan"] = {
        "h_range": [14, 22, 0.2],
        "h_mark": 18,
        "magtype": "asteroid",
        "char": "inner",
    }
    defaults["TNO"] = {
        "h_range": [4, 12, 0.2],
        "h_mark": 8,
        "magtype": "asteroid",
        "char": "outer",
    }
    defaults["SDO"] = {
        "h_range": [4, 12, 0.2],
        "h_mark": 8,
        "magtype": "asteroid",
        "char": "outer",
    }
    defaults["LPC"] = {
        "h_range": [6, 22, 0.5],
        "h_mark": 13,
        "magtype": "comet_oort",
        "char": "outer",
    }
    defaults["SPC"] = {
        "h_range": [4, 20, 0.5],
        "h_mark": 8,
        "magtype": "comet_short",
        "char": "outer",
    }
    defaults["generic"] = {
        "h_range": [4, 28, 0.5],
        "h_mark": 10,
        "magtype": "asteroid",
        "char": "inner",
    }
    # Some objtypes may be provided by other names
    if objtype.upper().startswith("GRANVIK"):
        objtype = "NEO"
    if objtype.upper().startswith("L7"):
        objtype = "TNO"
    if objtype.upper().startswith("OCC"):
        objtype = "LPC"

    if objtype not in defaults:
        print(
            f"## Could not find {objtype} in default keys ({defaults.keys()}). \n"
            f"## Using generic default values instead."
        )
        objtype = "generic"

    return defaults[objtype]


def quick_discovery_batch(
    slicer,
    colmap=None,
    run_name="run_name",
    detection_losses="detection",
    objtype="",
    albedo=None,
    h_mark=None,
    np_reduce=np.mean,
    constraint_info_label="",
    constraint=None,
    magtype="asteroid",
):
    """A subset of discovery metrics, using only the default discovery
    criteria of 3 pairs in 15 or 30 nights.
    """
    if colmap is None:
        colmap = col_map_dict()
    bundleList = []

    basicPlotDict = {
        "albedo": albedo,
        "h_mark": h_mark,
        "np_reduce": np_reduce,
        "nxbins": 200,
        "nybins": 200,
    }
    plot_funcs = [plots.MetricVsH()]
    display_dict = {"group": f"{objtype}", "subgroup": "Discovery"}

    if constraint_info_label == "" and constraint is not None:
        constraint_info_label = constraint.replace("filter", "").replace("==", "").replace("  ", " ")
    info_label = objtype + " " + constraint_info_label
    info_label = info_label.rstrip(" ")

    if detection_losses not in ("detection", "trailing"):
        raise ValueError("Please choose detection or trailing as options for detection_losses.")
    if detection_losses == "trailing":
        magStacker = stackers.MoMagStacker(loss_col="dmag_trail", magtype=magtype)
        detection_losses = " trailing loss"
    else:
        magStacker = stackers.MoMagStacker(loss_col="dmag_detect", magtype=magtype)
        detection_losses = " detection loss"

    # Set up a dictionary to pass to each metric for the column names.
    colkwargs = {
        "mjd_col": colmap["mjd"],
        "seeing_col": colmap["seeingGeom"],
        "exp_time_col": colmap["exptime"],
        "m5_col": colmap["fiveSigmaDepth"],
        "night_col": colmap["night"],
        "filter_col": colmap["filter"],
    }

    def _setup_child_metrics(parentMetric):
        childMetrics = {}
        childMetrics["Time"] = metrics.DiscoveryTimeMetric(parentMetric, **colkwargs)
        childMetrics["N_Chances"] = metrics.DiscoveryNChancesMetric(parentMetric, **colkwargs)
        # Could expand to add N_chances per year, but not really necessary.
        return childMetrics

    def _configure_child_bundles(parentBundle):
        dispDict = {
            "group": f"{objtype}",
            "subgroup": "Completeness Over Time",
            "caption": "Time of discovery of objects",
            "order": 0,
        }
        parentBundle.child_bundles["Time"].set_display_dict(dispDict)
        dispDict = {
            "group": f"{objtype}",
            "subgroup": "N Chances",
            "caption": "Number of chances for discovery of objects",
            "order": 0,
        }
        parentBundle.child_bundles["N_Chances"].set_display_dict(dispDict)
        return

    t_min = 5.0 / 60.0 / 24.0
    t_max = 90.0 / 60.0 / 24.0

    # 3 pairs in 15
    md = info_label + " 3 pairs in 15 nights" + detection_losses
    # Set up plot dict.
    plotDict = {"title": "%s: %s" % (run_name, md)}
    plotDict.update(basicPlotDict)
    metric = metrics.DiscoveryMetric(
        n_obs_per_night=2,
        t_min=t_min,
        t_max=t_max,
        n_nights_per_window=3,
        t_window=15,
        **colkwargs,
    )
    childMetrics = _setup_child_metrics(metric)
    bundle = MoMetricBundle(
        metric,
        slicer,
        constraint,
        stacker_list=[magStacker],
        run_name=run_name,
        info_label=md,
        child_metrics=childMetrics,
        plot_dict=plotDict,
        plot_funcs=plot_funcs,
        display_dict=display_dict,
    )
    _configure_child_bundles(bundle)
    bundleList.append(bundle)

    # 3 pairs in 30
    md = info_label + " 3 pairs in 30 nights" + detection_losses
    plotDict = {"title": "%s: %s" % (run_name, md)}
    plotDict.update(basicPlotDict)
    metric = metrics.DiscoveryMetric(
        n_obs_per_night=2,
        t_min=t_min,
        t_max=t_max,
        n_nights_per_window=3,
        t_window=30,
        **colkwargs,
    )
    childMetrics = _setup_child_metrics(metric)
    bundle = MoMetricBundle(
        metric,
        slicer,
        constraint,
        stacker_list=[magStacker],
        run_name=run_name,
        info_label=md,
        child_metrics=childMetrics,
        plot_dict=plotDict,
        plot_funcs=plot_funcs,
        display_dict=display_dict,
    )
    _configure_child_bundles(bundle)
    bundleList.append(bundle)

    # Set the run_name for all bundles and return the bundleDict.
    for b in bundleList:
        b.set_run_name(run_name)
    return mb.make_bundles_dict_from_list(bundleList)


def discovery_batch(
    slicer,
    colmap=None,
    run_name="run_name",
    detection_losses="detection",
    objtype="",
    albedo=None,
    h_mark=None,
    np_reduce=np.mean,
    constraint_info_label="",
    constraint=None,
    magtype="asteroid",
):
    """A comprehensive set of discovery metrics, using a wide range
    of discovery criteria.
    """
    if colmap is None:
        colmap = col_map_dict()
    bundleList = []

    basicPlotDict = {
        "albedo": albedo,
        "h_mark": h_mark,
        "np_reduce": np_reduce,
        "nxbins": 200,
        "nybins": 200,
    }
    plot_funcs = [plots.MetricVsH()]
    display_dict = {"group": f"{objtype}", "subgroup": "Discovery"}

    if constraint_info_label == "" and constraint is not None:
        constraint_info_label = constraint.replace("filter", "").replace("==", "").replace("  ", " ")
    info_label = objtype + " " + constraint_info_label
    info_label = info_label.rstrip(" ")

    if detection_losses not in ("detection", "trailing"):
        raise ValueError("Please choose detection or trailing as options for detection_losses.")
    if detection_losses == "trailing":
        # These are the SNR-losses only.
        magStacker = stackers.MoMagStacker(loss_col="dmag_trail", magtype=magtype)
        detection_losses = " trailing loss"
    else:
        # SNR losses, plus additional loss due to detecting with stellar PSF.
        magStacker = stackers.MoMagStacker(loss_col="dmag_detect", magtype=magtype)
        detection_losses = " detection loss"

    # Set up a dictionary to pass to each metric for the column names.
    colkwargs = {
        "mjd_col": colmap["mjd"],
        "seeing_col": colmap["seeingGeom"],
        "exp_time_col": colmap["exptime"],
        "m5_col": colmap["fiveSigmaDepth"],
        "night_col": colmap["night"],
        "filter_col": colmap["filter"],
    }

    def _setup_child_metrics(parentMetric):
        childMetrics = {}
        childMetrics["Time"] = metrics.DiscoveryTimeMetric(parentMetric, **colkwargs)
        childMetrics["N_Chances"] = metrics.DiscoveryNChancesMetric(parentMetric, **colkwargs)
        # Could expand to add N_chances per year, but not really necessary.
        return childMetrics

    def _configure_child_bundles(parentBundle):
        dispDict = {
            "group": f"{objtype}",
            "subgroup": "Completeness Over Time",
            "caption": "Time of discovery of objects",
            "order": 0,
        }
        parentBundle.child_bundles["Time"].set_display_dict(dispDict)
        dispDict = {
            "group": f"{objtype}",
            "subgroup": "N Chances",
            "caption": "Number of chances for discovery of objects",
            "order": 0,
        }
        parentBundle.child_bundles["N_Chances"].set_display_dict(dispDict)

    t_min = 5.0 / 60.0 / 24.0
    t_max = 90.0 / 60.0 / 24.0

    # 3 pairs in 15 and 3 pairs in 30 done in 'quickDiscoveryBatch' (with vis).

    # 4 pairs in 20
    md = info_label + " 4 pairs in 20 nights" + detection_losses
    plotDict = {"title": "%s: %s" % (run_name, md)}
    plotDict.update(basicPlotDict)
    metric = metrics.DiscoveryMetric(
        n_obs_per_night=2,
        t_min=t_min,
        t_max=t_max,
        n_nights_per_window=4,
        t_window=20,
        **colkwargs,
    )
    childMetrics = _setup_child_metrics(metric)
    bundle = MoMetricBundle(
        metric,
        slicer,
        constraint,
        stacker_list=[magStacker],
        run_name=run_name,
        info_label=md,
        child_metrics=childMetrics,
        plot_dict=plotDict,
        plot_funcs=plot_funcs,
        display_dict=display_dict,
    )
    _configure_child_bundles(bundle)
    bundleList.append(bundle)

    # 3 triplets in 30
    md = info_label + " 3 triplets in 30 nights" + detection_losses
    plotDict = {"title": "%s: %s" % (run_name, md)}
    plotDict.update(basicPlotDict)
    metric = metrics.DiscoveryMetric(
        n_obs_per_night=3,
        t_min=t_min,
        t_max=120.0 / 60.0 / 24.0,
        n_nights_per_window=3,
        t_window=30,
        **colkwargs,
    )
    childMetrics = _setup_child_metrics(metric)
    bundle = MoMetricBundle(
        metric,
        slicer,
        constraint,
        stacker_list=[magStacker],
        run_name=run_name,
        info_label=md,
        child_metrics=childMetrics,
        plot_dict=plotDict,
        plot_funcs=plot_funcs,
        display_dict=display_dict,
    )
    _configure_child_bundles(bundle)
    bundleList.append(bundle)

    # 1 quad
    md = info_label + " 1 quad in 1 night" + detection_losses
    plotDict = {"title": "%s: %s" % (run_name, md)}
    plotDict.update(basicPlotDict)
    metric = metrics.DiscoveryMetric(
        n_obs_per_night=4,
        t_min=t_min,
        t_max=150.0 / 60.0 / 24.0,
        n_nights_per_window=1,
        t_window=2,
        **colkwargs,
    )
    childMetrics = _setup_child_metrics(metric)
    bundle = MoMetricBundle(
        metric,
        slicer,
        constraint,
        stacker_list=[magStacker],
        run_name=run_name,
        info_label=md,
        child_metrics=childMetrics,
        plot_dict=plotDict,
        plot_funcs=plot_funcs,
        display_dict=display_dict,
    )
    _configure_child_bundles(bundle)
    bundleList.append(bundle)

    # Play with SNR.
    # First standard SNR / probabilistic visibility (SNR~5)
    # 3 pairs in 15
    md = info_label + " 3 pairs in 15 nights SNR=5" + detection_losses
    # Set up plot dict.
    plotDict = {"title": "%s: %s" % (run_name, md)}
    plotDict.update(basicPlotDict)
    metric = metrics.DiscoveryMetric(
        n_obs_per_night=2,
        t_min=t_min,
        t_max=t_max,
        n_nights_per_window=3,
        t_window=15,
        snr_limit=5,
        **colkwargs,
    )
    childMetrics = _setup_child_metrics(metric)
    bundle = MoMetricBundle(
        metric,
        slicer,
        constraint,
        stacker_list=[magStacker],
        run_name=run_name,
        info_label=md,
        child_metrics=childMetrics,
        plot_dict=plotDict,
        plot_funcs=plot_funcs,
        display_dict=display_dict,
    )
    _configure_child_bundles(bundle)
    bundleList.append(bundle)

    # 3 pairs in 15, SNR=4.
    md = info_label + " 3 pairs in 15 nights SNR=4" + detection_losses
    plotDict = {"title": "%s: %s" % (run_name, md)}
    plotDict.update(basicPlotDict)
    metric = metrics.DiscoveryMetric(
        n_obs_per_night=2,
        t_min=t_min,
        t_max=t_max,
        n_nights_per_window=3,
        t_window=15,
        snr_limit=4,
        **colkwargs,
    )
    childMetrics = _setup_child_metrics(metric)
    bundle = MoMetricBundle(
        metric,
        slicer,
        constraint,
        stacker_list=[magStacker],
        run_name=run_name,
        info_label=md,
        child_metrics=childMetrics,
        plot_dict=plotDict,
        plot_funcs=plot_funcs,
        display_dict=display_dict,
    )
    _configure_child_bundles(bundle)
    bundleList.append(bundle)

    # 3 pairs in 15, SNR=3
    md = info_label + " 3 pairs in 15 nights SNR=3" + detection_losses
    plotDict = {"title": "%s: %s" % (run_name, md)}
    plotDict.update(basicPlotDict)
    metric = metrics.DiscoveryMetric(
        n_obs_per_night=2,
        t_min=t_min,
        t_max=t_max,
        n_nights_per_window=3,
        t_window=15,
        snr_limit=3,
        **colkwargs,
    )
    childMetrics = _setup_child_metrics(metric)
    bundle = MoMetricBundle(
        metric,
        slicer,
        constraint,
        stacker_list=[magStacker],
        run_name=run_name,
        info_label=md,
        child_metrics=childMetrics,
        plot_dict=plotDict,
        plot_funcs=plot_funcs,
        display_dict=display_dict,
    )
    _configure_child_bundles(bundle)
    bundleList.append(bundle)

    # SNR = 0
    # 3 pairs in 15, SNR=0
    md = info_label + " 3 pairs in 15 nights SNR=0" + detection_losses
    plotDict = {"title": "%s: %s" % (run_name, md)}
    plotDict.update(basicPlotDict)
    metric = metrics.DiscoveryMetric(
        n_obs_per_night=2,
        t_min=t_min,
        t_max=t_max,
        n_nights_per_window=3,
        t_window=15,
        snr_limit=0,
        **colkwargs,
    )
    childMetrics = _setup_child_metrics(metric)
    bundle = MoMetricBundle(
        metric,
        slicer,
        constraint,
        stacker_list=[magStacker],
        run_name=run_name,
        info_label=md,
        child_metrics=childMetrics,
        plot_dict=plotDict,
        plot_funcs=plot_funcs,
        display_dict=display_dict,
    )
    _configure_child_bundles(bundle)
    bundleList.append(bundle)

    # Play with weird strategies.
    # Single detection.
    md = info_label + " Single detection" + detection_losses
    plotDict = {"title": "%s: %s" % (run_name, md)}
    plotDict.update(basicPlotDict)
    metric = metrics.DiscoveryMetric(
        n_obs_per_night=1,
        t_min=t_min,
        t_max=t_max,
        n_nights_per_window=1,
        t_window=5,
        **colkwargs,
    )
    childMetrics = _setup_child_metrics(metric)
    bundle = MoMetricBundle(
        metric,
        slicer,
        constraint,
        stacker_list=[magStacker],
        run_name=run_name,
        info_label=md,
        child_metrics=childMetrics,
        plot_dict=plotDict,
        plot_funcs=plot_funcs,
        display_dict=display_dict,
    )
    _configure_child_bundles(bundle)
    bundleList.append(bundle)

    # Single pair of detections.
    md = info_label + " Single pair" + detection_losses
    plotDict = {"title": "%s: %s" % (run_name, md)}
    plotDict.update(basicPlotDict)
    metric = metrics.DiscoveryMetric(
        n_obs_per_night=2,
        t_min=t_min,
        t_max=t_max,
        n_nights_per_window=1,
        t_window=5,
        **colkwargs,
    )
    childMetrics = _setup_child_metrics(metric)
    bundle = MoMetricBundle(
        metric,
        slicer,
        constraint,
        stacker_list=[magStacker],
        run_name=run_name,
        info_label=md,
        child_metrics=childMetrics,
        plot_dict=plotDict,
        plot_funcs=plot_funcs,
        display_dict=display_dict,
    )
    _configure_child_bundles(bundle)
    bundleList.append(bundle)

    # High velocity discovery.
    md = info_label + " High velocity pair" + detection_losses
    plotDict = {"title": "%s: %s" % (run_name, md)}
    plotDict.update(basicPlotDict)
    metric = metrics.HighVelocityNightsMetric(psf_factor=2.0, n_obs_per_night=2, **colkwargs)
    bundle = MoMetricBundle(
        metric,
        slicer,
        constraint,
        stacker_list=[magStacker],
        run_name=run_name,
        info_label=md,
        plot_dict=plotDict,
        plot_funcs=plot_funcs,
        display_dict=display_dict,
    )
    bundleList.append(bundle)

    # "magic" detection - 6 in 60 days.
    md = info_label + " 6 detections in 60 nights" + detection_losses
    plotDict = {"title": "%s: %s" % (run_name, md)}
    plotDict.update(basicPlotDict)
    metric = metrics.MagicDiscoveryMetric(n_obs=6, t_window=60, **colkwargs)
    bundle = MoMetricBundle(
        metric,
        slicer,
        constraint,
        stacker_list=[magStacker],
        run_name=run_name,
        info_label=md,
        plot_dict=plotDict,
        plot_funcs=plot_funcs,
        display_dict=display_dict,
    )
    bundleList.append(bundle)

    # Set the run_name for all bundles and return the bundleDict.
    for b in bundleList:
        b.set_run_name(run_name)
    return mb.make_bundles_dict_from_list(bundleList)


def run_completeness_summary(bdict, h_mark, times, out_dir, results_db):
    """Calculate completeness and create completeness bundles from all
    N_Chances and Time (child) metrics of the (discovery) bundles in bdict,
    and write completeness at h_mark to results_db, save bundle to disk.

    This should be done after combining any sub-sets of the metric results.

    Parameters
    ----------
    bdict : `dict` of `maf.MetricBundle`
        Dict containing ~rubin_sim.maf.MoMetricBundles,
        including bundles we're expecting to contain completeness.
    h_mark : `float`
        h_mark value to add to completeness plotting dict.
        If not defined (None), then the h_mark from the plotdict from the
        metric bundle will be used if available.
        If None and h_mark not in plot_dict, then median of h_range values
        will be used.
    times : `np.ndarray`
        The times at which to calculate completeness (over time).
    out_dir : `str`
        Output directory to save completeness bundles to disk.
    results_db : `maf.ResultsDb`
        Results database to save information about completeness bundle.

    Returns
    -------
    metricDict : `dict` of `maf.MetricBundles`
        A dictionary of the new completeness bundles. Keys match original keys,
        with additions of "[Differential,Cumulative]Completeness@Time"
        and "[Differential,Cumulative]Completeness" to distinguish new entries.
    """
    # Add completeness bundles and write completeness at h_mark to results_db.
    completeness = {}

    def _compbundles(b, bundle, h_mark, results_db):
        # Find h_mark if not set (this may be different for different bundles).
        if h_mark is None and "h_mark" in bundle.plot_dict:
            h_mark = bundle.plot_dict["h_mark"]
        if h_mark is None:
            h_mark = np.median(bundle.slicer.slice_points["H"])
        # Set up the summary metrics.
        summaryTimeMetrics = summary_completeness_at_time(times, h_val=h_mark, h_index=0.33)
        summaryTimeMetrics2 = summary_completeness_at_time(times, h_val=h_mark - 2, h_index=0.33)
        summaryHMetrics = summary_completeness_over_h(requiredChances=1, Hindex=0.33)
        comp = {}
        # Bundle = single metric bundle.
        # Add differential and cumulative completeness.
        if "Time" in bundle.metric.name:
            for metric in summaryTimeMetrics:
                newkey = b + " " + metric.name
                comp[newkey] = mb.make_completeness_bundle(bundle, metric, h_mark=None, results_db=results_db)
                comp[newkey].plot_dict["times"] = times
                comp[newkey].plot_dict["h_val"] = metric.hval
            for metric in summaryTimeMetrics2:
                newkey = b + " " + metric.name
                comp[newkey] = mb.make_completeness_bundle(bundle, metric, h_mark=None, results_db=results_db)
                comp[newkey].plot_dict["times"] = times
                comp[newkey].plot_dict["h_val"] = metric.hval
        elif "NChances" in bundle.metric.name or "N_Chances" in bundle.metric.name:
            for metric in summaryHMetrics:
                newkey = b + " " + metric.name
                comp[newkey] = mb.make_completeness_bundle(
                    bundle, metric, h_mark=h_mark, results_db=results_db
                )
        elif "MagicDiscovery" in bundle.metric.name:
            for metric in summaryHMetrics:
                newkey = b + " " + metric.name
                comp[newkey] = mb.make_completeness_bundle(
                    bundle, metric, h_mark=h_mark, results_db=results_db
                )
        elif "HighVelocity" in bundle.metric.name:
            for metric in summaryHMetrics:
                newkey = b + " " + metric.name
                comp[newkey] = mb.make_completeness_bundle(
                    bundle, metric, h_mark=h_mark, results_db=results_db
                )
        return comp

    # Generate the completeness bundles for the various discovery metrics.
    for b, bundle in bdict.items():
        if "Discovery" in bundle.metric.name:
            completeness.update(_compbundles(b, bundle, h_mark, results_db))
        if "MagicDiscovery" in bundle.metric.name:
            completeness.update(_compbundles(b, bundle, h_mark, results_db))
        if "HighVelocity" in bundle.metric.name:
            completeness.update(_compbundles(b, bundle, h_mark, results_db))
    # Write the completeness bundles to disk, so we can re-read them later.
    # (also set the display dict properties, for the results_db output).
    for b, bundle in completeness.items():
        bundle.display_dict["subgroup"] = "Completeness"
        bundle.write(out_dir=out_dir, results_db=results_db)

    # Calculate total number of objects - currently for NEOs and PHAs only
    for b, bundle in completeness.items():
        if "DifferentialCompleteness" in b and "@Time" not in b:
            if "NEO" in bundle.info_label:
                nobj_metrics = [
                    metrics.TotalNumberSSO(h_mark=22, dndh_func=metrics.neo_dndh_granvik),
                    metrics.TotalNumberSSO(h_mark=25, dndh_func=metrics.neo_dndh_granvik),
                ]
                bundle.set_summary_metrics(nobj_metrics)
                bundle.compute_summary_stats(results_db)
            if "PHA" in bundle.info_label:
                nobj_metrics = [metrics.TotalNumberSSO(h_mark=22, dndh_func=metrics.pha_dndh_granvik)]
                bundle.set_summary_metrics(nobj_metrics)
                bundle.compute_summary_stats(results_db)
    return completeness


def plot_completeness(
    bdictCompleteness,
    figroot=None,
    run_name=None,
    results_db=None,
    out_dir=".",
    fig_format="pdf",
):
    """Plot a minor subset of the completeness results."""
    # Separate some subsets to plot together -
    # first just the simple 15 and 30 night detection loss metrics.
    keys = [
        "3_pairs_in_30_nights_detection_loss",
        "3_pairs_in_15_nights_detection_loss",
    ]
    plotTimes = {}
    plotComp = {}
    plotDiff = {}
    for k in bdictCompleteness:
        for key in keys:
            if key in k:
                if "Time" in k:
                    if "Cumulative" in k:
                        plotTimes[k] = bdictCompleteness[k]
                elif "Chances" in k:
                    if "Differential" in k:
                        plotDiff[k] = bdictCompleteness[k]
                    elif "Cumulative" in k:
                        plotComp[k] = bdictCompleteness[k]

    # Add plot dictionaries to code 30 nights red, 15 nights blue,
    # differentials dotted.
    def _codePlot(key):
        plotDict = {}
        if "Differential" in k:
            plotDict["linestyle"] = ":"
        else:
            plotDict["linestyle"] = "-"
        if "30_nights" in k:
            plotDict["color"] = "r"
        if "15_nights" in k:
            plotDict["color"] = "b"
        return plotDict

    # Apply color-coding.
    for k, b in plotTimes.items():
        b.set_plot_dict(_codePlot(k))
    for k, b in plotDiff.items():
        b.set_plot_dict(_codePlot(k))
    for k, b in plotComp.items():
        b.set_plot_dict(_codePlot(k))

    first = bdictCompleteness[list(bdictCompleteness.keys())[0]]
    if run_name is None:
        run_name = first.run_name
    if figroot is None:
        figroot = run_name
    display_dict = deepcopy(first.display_dict)

    # Plot completeness as a function of time.
    # Make custom plot, then save it with PlotHandler.
    fig = plt.figure(figsize=(8, 6))
    for k in plotTimes:
        plt.plot(
            plotTimes[k].plot_dict["times"],
            plotTimes[k].metric_values[0, :],
            label=plotTimes[k].plot_dict["label"] + " @H=%.2f" % plotTimes[k].plot_dict["h_val"],
        )
    plt.legend()
    plt.xlabel("Time (MJD)")
    plt.ylabel("Completeness")
    plt.grid(True, alpha=0.3)
    # Make a PlotHandler to deal with savings/results_db, etc.
    ph = plots.PlotHandler(fig_format=fig_format, results_db=results_db, out_dir=out_dir)
    display_dict["subgroup"] = "Completeness over time"
    display_dict["caption"] = "Completeness over time, for H values indicated in legend."
    ph.save_fig(
        fig,
        f"{figroot}_CompletenessOverTime",
        "Combo",
        "CompletenessOverTime",
        "MoObjSlicer",
        figroot,
        None,
        None,
        display_dict=display_dict,
    )

    plt.savefig(
        os.path.join(out_dir, f"{figroot}_CompletenessOverTime.{fig_format}"),
        format=fig_format,
    )

    # Plot cumulative completeness.
    ph = plots.PlotHandler(fig_format=fig_format, results_db=results_db, out_dir=out_dir)
    ph.set_metric_bundles(plotComp)
    plotDict = {"ylabel": "Completeness", "figsize": (8, 6), "albedo": 0.14}
    ph.plot(
        plot_func=plots.MetricVsH(),
        plot_dicts=plotDict,
        outfile_root=figroot + "_CumulativeCompleteness",
    )

    # Plot differential completeness.
    ph = plots.PlotHandler(fig_format=fig_format, results_db=results_db, out_dir=out_dir)
    ph.set_metric_bundles(plotDiff)
    plotDict = {"ylabel": "Completeness", "figsize": (8, 6)}
    ph.plot(
        plot_func=plots.MetricVsH(),
        plot_dicts=plotDict,
        outfile_root=figroot + "_DifferentialCompleteness",
    )

    # And add the rest of the completeness calculations.
    allComp = []
    for k in bdictCompleteness:
        if "DiscoveryNChances" in k:
            if "Cumulative" in k:
                allComp.append(bdictCompleteness[k])
        if "Magic" in k:
            if "Cumulative" in k:
                allComp.append(bdictCompleteness[k])
    ph = plots.PlotHandler(fig_format=fig_format, results_db=results_db, out_dir=out_dir)
    ph.set_metric_bundles(allComp)
    plotDict = {
        "ylabel": "Completeness",
        "figsize": (8, 6),
        "legendloc": (1.01, 0.1),
        "color": None,
    }
    display_dict["subgroup"] = "Completeness all criteria"
    display_dict["caption"] = "Plotting all of the cumulative completeness curves together."
    ph.plot(
        plot_func=plots.MetricVsH(),
        plot_dicts=plotDict,
        display_dict=display_dict,
        outfile_root=figroot + "_Many_CumulativeCompleteness",
    )


def characterization_inner_batch(
    slicer,
    colmap=None,
    run_name="run_name",
    objtype="",
    magtype="asteroid",
    albedo=None,
    h_mark=None,
    constraint_info_label="",
    constraint=None,
    npReduce=np.mean,
    windows=None,
    bins=None,
):
    """Characterization metrics for inner solar system objects."""
    if colmap is None:
        colmap = col_map_dict()
    bundleList = []

    # Set up a dictionary to pass to each metric for the column names.
    colkwargs = {
        "mjd_col": colmap["mjd"],
        "seeing_col": colmap["seeingGeom"],
        "exp_time_col": colmap["exptime"],
        "m5_col": colmap["fiveSigmaDepth"],
        "night_col": colmap["night"],
        "filter_col": colmap["filter"],
    }

    basicPlotDict = {
        "albedo": albedo,
        "h_mark": h_mark,
        "np_reduce": npReduce,
        "nxbins": 200,
        "nybins": 200,
    }
    plot_funcs = [plots.MetricVsH()]

    if constraint_info_label == "" and constraint is not None:
        constraint_info_label = constraint.replace("filter", "").replace("==", "").replace("  ", " ")
    info_label = objtype + " " + constraint_info_label
    info_label = info_label.rstrip(" ")

    display_dict = {"group": f"{objtype}"}

    # Stackers
    magStacker = stackers.MoMagStacker(loss_col="dmag_detect", magtype=magtype)
    eclStacker = stackers.EclStacker()
    stackerList = [magStacker, eclStacker]

    # Windows are the different 'length of activity'
    if windows is None:
        windows = np.arange(10, 200, 30.0)
    # Bins are the different 'anomaly variations' of activity
    if bins is None:
        bins = np.arange(5, 185, 20.0)

    # Number of observations.
    md = info_label
    display_dict["subgroup"] = "N Obs"
    plotDict = {
        "ylabel": "Number of observations (#)",
        "title": "%s: Number of observations %s" % (run_name, md),
    }
    plotDict.update(basicPlotDict)
    metric = metrics.NObsMetric(**colkwargs)
    bundle = MoMetricBundle(
        metric,
        slicer,
        constraint,
        stacker_list=stackerList,
        run_name=run_name,
        info_label=md,
        plot_dict=plotDict,
        plot_funcs=plot_funcs,
        display_dict=display_dict,
    )
    bundleList.append(bundle)

    # Observational arc.
    md = info_label
    display_dict["subgroup"] = "Obs Arc"
    plotDict = {
        "ylabel": "Observational Arc (days)",
        "title": "%s: Observational Arc Length %s" % (run_name, md),
    }
    plotDict.update(basicPlotDict)
    metric = metrics.ObsArcMetric(**colkwargs)
    bundle = MoMetricBundle(
        metric,
        slicer,
        constraint,
        stacker_list=stackerList,
        run_name=run_name,
        info_label=md,
        plot_dict=plotDict,
        plot_funcs=plot_funcs,
        display_dict=display_dict,
    )
    bundleList.append(bundle)

    # Activity detection.
    display_dict["subgroup"] = "Activity"
    for w in windows:
        md = info_label + " activity lasting %.0f days" % w
        plotDict = {
            "title": "%s: Chances of detecting %s" % (run_name, md),
            "ylabel": "Probability of detection per %.0f day window" % w,
        }
        metric_name = "Chances of detecting activity lasting %.0f days" % w
        metric = metrics.ActivityOverTimeMetric(w, metric_name=metric_name, **colkwargs)
        bundle = MoMetricBundle(
            metric,
            slicer,
            constraint,
            stacker_list=stackerList,
            run_name=run_name,
            info_label=info_label,
            plot_dict=plotDict,
            plot_funcs=plot_funcs,
            display_dict=display_dict,
        )
        bundleList.append(bundle)

    for b in bins:
        md = info_label + " activity covering %.0f deg" % (b)
        plotDict = {
            "title": "%s: Chances of detecting %s" % (run_name, md),
            "ylabel": "Probability of detection per %.0f deg window" % b,
        }
        metric_name = "Chances of detecting activity covering %.0f deg" % (b)
        metric = metrics.ActivityOverPeriodMetric(b, metric_name=metric_name, **colkwargs)
        bundle = MoMetricBundle(
            metric,
            slicer,
            constraint,
            stacker_list=stackerList,
            run_name=run_name,
            info_label=info_label,
            plot_dict=plotDict,
            plot_funcs=plot_funcs,
            display_dict=display_dict,
        )
        bundleList.append(bundle)

    # Lightcurve inversion.
    md = info_label
    display_dict["subgroup"] = "Color/Inversion"
    plotDict = {
        "y_min": 0,
        "y_max": 1,
        "ylabel": "Fraction of objects",
        "title": "%s: Fraction with potential lightcurve inversion %s" % (run_name, md),
    }
    plotDict.update(basicPlotDict)
    metric = metrics.LightcurveInversionAsteroidMetric(**colkwargs)
    bundle = MoMetricBundle(
        metric,
        slicer,
        constraint,
        stacker_list=stackerList,
        run_name=run_name,
        info_label=md,
        plot_dict=plotDict,
        plot_funcs=plot_funcs,
        display_dict=display_dict,
    )
    bundleList.append(bundle)

    # Color determination.
    md = info_label
    plotDict = {
        "y_min": 0,
        "y_max": 1,
        "ylabel": "Fraction of objects",
        "title": "%s: Fraction of population with colors in X filters %s" % (run_name, md),
    }
    plotDict.update(basicPlotDict)
    metric = metrics.ColorAsteroidMetric(**colkwargs)
    bundle = MoMetricBundle(
        metric,
        slicer,
        constraint,
        stacker_list=stackerList,
        run_name=run_name,
        info_label=md,
        plot_dict=plotDict,
        plot_funcs=plot_funcs,
        display_dict=display_dict,
    )
    bundleList.append(bundle)

    # Set the run_name for all bundles and return the bundleDict.
    for b in bundleList:
        b.set_run_name(run_name)
    return mb.make_bundles_dict_from_list(bundleList)


def characterization_outer_batch(
    slicer,
    colmap=None,
    run_name="run_name",
    objtype="",
    magtype="asteroid",
    albedo=None,
    h_mark=None,
    constraint_info_label="",
    constraint=None,
    npReduce=np.mean,
    windows=None,
    bins=None,
):
    """Characterization metrics for outer solar system objects."""
    if colmap is None:
        colmap = col_map_dict()
    bundleList = []

    # Set up a dictionary to pass to each metric for the column names.
    colkwargs = {
        "mjd_col": colmap["mjd"],
        "seeing_col": colmap["seeingGeom"],
        "exp_time_col": colmap["exptime"],
        "m5_col": colmap["fiveSigmaDepth"],
        "night_col": colmap["night"],
        "filter_col": colmap["filter"],
    }

    basicPlotDict = {
        "albedo": albedo,
        "h_mark": h_mark,
        "np_reduce": npReduce,
        "nxbins": 200,
        "nybins": 200,
    }
    plot_funcs = [plots.MetricVsH()]

    if constraint_info_label == "" and constraint is not None:
        constraint_info_label = constraint.replace("filter", "").replace("==", "").replace("  ", " ")
    info_label = objtype + " " + constraint_info_label
    info_label = info_label.rstrip(" ")

    display_dict = {"group": f"{objtype}"}

    # Stackers
    magStacker = stackers.MoMagStacker(loss_col="dmag_detect", magtype=magtype)
    eclStacker = stackers.EclStacker()
    stackerList = [magStacker, eclStacker]

    # Windows are the different 'length of activity'
    if windows is None:
        windows = np.arange(10, 200, 30.0)
    # Bins are the different 'anomaly variations' of activity
    if bins is None:
        bins = np.arange(5, 185, 20.0)

    # Number of observations.
    md = info_label
    display_dict["subgroup"] = "N Obs"
    plotDict = {
        "ylabel": "Number of observations (#)",
        "title": "%s: Number of observations %s" % (run_name, md),
    }
    plotDict.update(basicPlotDict)
    metric = metrics.NObsMetric(**colkwargs)
    bundle = MoMetricBundle(
        metric,
        slicer,
        constraint,
        stacker_list=stackerList,
        run_name=run_name,
        info_label=md,
        plot_dict=plotDict,
        plot_funcs=plot_funcs,
        display_dict=display_dict,
    )
    bundleList.append(bundle)

    # Observational arc.
    md = info_label
    display_dict["subgroup"] = "Obs Arc"
    plotDict = {
        "ylabel": "Observational Arc (days)",
        "title": "%s: Observational Arc Length %s" % (run_name, md),
    }
    plotDict.update(basicPlotDict)
    metric = metrics.ObsArcMetric(**colkwargs)
    bundle = MoMetricBundle(
        metric,
        slicer,
        constraint,
        stacker_list=stackerList,
        run_name=run_name,
        info_label=md,
        plot_dict=plotDict,
        plot_funcs=plot_funcs,
        display_dict=display_dict,
    )
    bundleList.append(bundle)

    # Activity detection.
    display_dict["subgroup"] = "Activity"
    for w in windows:
        md = info_label + " activity lasting %.0f days" % w
        plotDict = {
            "title": "%s: Chances of detecting %s" % (run_name, md),
            "ylabel": "Probability of detection per %.0f day window" % w,
        }
        metric_name = "Chances of detecting activity lasting %.0f days" % w
        metric = metrics.ActivityOverTimeMetric(w, metric_name=metric_name, **colkwargs)
        bundle = MoMetricBundle(
            metric,
            slicer,
            constraint,
            stacker_list=stackerList,
            run_name=run_name,
            info_label=info_label,
            plot_dict=plotDict,
            plot_funcs=plot_funcs,
            display_dict=display_dict,
        )
        bundleList.append(bundle)

    for b in bins:
        md = info_label + " activity covering %.0f deg" % (b)
        plotDict = {
            "title": "%s: Chances of detecting %s" % (run_name, md),
            "ylabel": "Probability of detection per %.2f deg window" % b,
        }
        metric_name = "Chances of detecting activity covering %.0f deg" % (b)
        metric = metrics.ActivityOverPeriodMetric(b, metric_name=metric_name, **colkwargs)
        bundle = MoMetricBundle(
            metric,
            slicer,
            constraint,
            stacker_list=stackerList,
            run_name=run_name,
            info_label=info_label,
            plot_dict=plotDict,
            plot_funcs=plot_funcs,
            display_dict=display_dict,
        )
        bundleList.append(bundle)

    # Color determination.
    md = info_label
    display_dict["subgroup"] = "Color/Inversion"
    plotDict = {
        "y_min": 0,
        "y_max": 1,
        "ylabel": "Fraction of objects",
        "title": "%s: Fraction of population with colors in X filters %s" % (run_name, md),
    }
    plotDict.update(basicPlotDict)
    metric = metrics.LightcurveColorOuterMetric(**colkwargs)
    bundle = MoMetricBundle(
        metric,
        slicer,
        constraint,
        stacker_list=stackerList,
        run_name=run_name,
        info_label=md,
        plot_dict=plotDict,
        plot_funcs=plot_funcs,
        display_dict=display_dict,
    )
    bundleList.append(bundle)

    # Set the run_name for all bundles and return the bundleDict.
    for b in bundleList:
        b.set_run_name(run_name)
    return mb.make_bundles_dict_from_list(bundleList)


def run_fraction_summary(bdict, h_mark, out_dir, results_db):
    """
    Calculate fractional completeness of the population for
    color and lightcurve metrics.

    This should be done after combining any sub-sets of the metric results.

    Parameters
    ----------
    bdict : `dict` of `maf.MoMetricBundle`
        Dict containing bundles contianing lightcurve/color evaluations.
    h_mark : `float`
        h_mark value to add to completeness plotting dict.
        If defined, this value is used.
        If None, but h_mark in plot_dict for metric, then this value (-2) is
        used. If h_mark not in plotdict, then use the median h_range value-2.
    times : `np.ndarray`
        The times at which to calculate completeness (over time).
    out_dir : `str`
        Output directory to save completeness bundles to disk.
    results_db : `maf.ResultsDb`
        Results database to save information about completeness bundle.

    Returns
    -------
    metricDict : `dict` of `maf.MetricBundle`
        Dictionary of the metric bundles for the fractional evaluation
        of the population.
    """
    fractions = {}

    # Look for metrics from asteroid or outer solar system
    # color/lightcurve metrics.
    inversionSummary = fraction_population_at_threshold([1], ["Lightcurve Inversion"])
    asteroidColorSummary = fraction_population_at_threshold(
        [4, 3, 2, 1],
        ["6 of ugrizy", "5 of grizy", "4 of grizy", "2 of g, r or i, z or y"],
    )
    asteroidSummaryMetrics = {
        "LightcurveInversionAsteroid": inversionSummary,
        "LightcurveInversion_Asteroid": inversionSummary,
        "ColorAsteroid": asteroidColorSummary,
        "Color_Asteroid": asteroidColorSummary,
    }

    outerColorSummary = fraction_population_at_threshold(
        [6, 5, 4, 3, 2, 1],
        ["6 filters", "5 filters", "4 filters", "3 filters", "2 filters", "1 filters"],
    )
    outerSummaryMetrics = {
        "LightcurveColorOuter": outerColorSummary,
        "lightcurveColor_Outer": outerColorSummary,
    }

    for b, bundle in bdict.items():
        # Find h_mark if not set (this may be different for different bundles).
        if h_mark is None and "h_mark" in bundle.plot_dict:
            h_mark = bundle.plot_dict["h_mark"] - 2
        if h_mark is None:
            h_mark = np.median(bundle.slicer.slice_points["H"]) - 2
        # Make sure we didn't push h_mark outside the range of H values
        if h_mark < bundle.slicer.slice_points["H"].min():
            h_mark = bundle.slicer.slice_points["H"].min()
        for k in asteroidSummaryMetrics:
            if k in b:
                for summary_metric in asteroidSummaryMetrics[k]:
                    newkey = b + " " + summary_metric.name
                    fractions[newkey] = mb.make_completeness_bundle(
                        bundle, summary_metric, h_mark=h_mark, results_db=results_db
                    )
        for k in outerSummaryMetrics:
            if k in b:
                for summary_metric in outerSummaryMetrics[k]:
                    newkey = b + " " + summary_metric.name
                    fractions[newkey] = mb.make_completeness_bundle(
                        bundle, summary_metric, h_mark=h_mark, results_db=results_db
                    )

    # Write fractional populations bundles to disk, so we can re-read later.
    for b, bundle in fractions.items():
        bundle.write(out_dir=out_dir, results_db=results_db)
    return fractions


def plot_fractions(
    bdictFractions,
    figroot=None,
    run_name=None,
    results_db=None,
    out_dir=".",
    fig_format="pdf",
):
    # Set colors for the fractions.
    for b in bdictFractions.values():
        k = b.metric.name
        if "6" in k:
            b.plot_dict["color"] = "b"
        if "5" in k:
            b.plot_dict["color"] = "cyan"
        if "4" in k:
            b.plot_dict["color"] = "orange"
        if "2" in k:
            b.plot_dict["color"] = "r"
        if "1" in k:
            b.plot_dict["color"] = "magenta"
        if "Lightcurve Inversion" in k:
            b.plot_dict["color"] = "k"
            b.plot_dict["linestyle"] = ":"
            b.plot_dict["linewidth"] = 3

    first = bdictFractions[list(bdictFractions.keys())[0]]
    if run_name is None:
        run_name = first.run_name
    if figroot is None:
        figroot = run_name
    display_dict = deepcopy(first.display_dict)
    display_dict["subgroup"] = "Characterization Fraction"

    ph = plots.PlotHandler(fig_format=fig_format, results_db=results_db, out_dir=out_dir)
    ph.set_metric_bundles(bdictFractions)
    ph.joint_metric_names = "Fraction of population for colors or lightcurve inversion"
    plotDict = {"ylabel": "Fraction of population", "figsize": (8, 6)}
    ph.plot(
        plot_func=plots.MetricVsH(),
        plot_dicts=plotDict,
        display_dict=display_dict,
        outfile_root=figroot + "_characterization",
    )


def plot_single(bundle, results_db=None, out_dir=".", fig_format="pdf"):
    """Plot 5%/25%/50%/75%/95% iles for a metric value."""
    pDict = {
        "95%ile": {
            "color": "k",
            "linestyle": "--",
            "label": "95th %ile",
            "np_reduce": lambda x, axis: np.percentile(x, 95, axis=axis),
        },
        "75%ile": {
            "color": "magenta",
            "linestyle": ":",
            "label": "75th %ile",
            "np_reduce": lambda x, axis: np.percentile(x, 75, axis=axis),
        },
        "Median": {
            "color": "b",
            "linestyle": "-",
            "label": "Median",
            "np_reduce": lambda x, axis: np.median(x, axis=axis),
        },
        "Mean": {
            "color": "g",
            "linestyle": "--",
            "label": "Mean",
            "np_reduce": np.mean,
        },
        "25%ile": {
            "color": "magenta",
            "linestyle": ":",
            "label": "25th %ile",
            "np_reduce": lambda x, axis: np.percentile(x, 25, axis=axis),
        },
        "5%ile": {
            "color": "k",
            "linestyle": "--",
            "label": "5th %ile",
            "np_reduce": lambda x, axis: np.percentile(x, 5, axis=axis),
        },
    }
    ph = plots.PlotHandler(fig_format=fig_format, results_db=results_db, out_dir=out_dir)
    plot_bundles = []
    plot_dicts = []
    for percentile in pDict:
        plot_bundles.append(bundle)
        plot_dicts.append(pDict[percentile])
    plot_dicts[0].update({"figsize": (8, 6), "legendloc": "upper right", "y_min": 0})
    # Remove the h_mark line because these plots get complicated already.
    for r in plot_dicts:
        r["h_mark"] = None
    ph.set_metric_bundles(plot_bundles)
    ph.plot(
        plot_func=plots.MetricVsH(),
        plot_dicts=plot_dicts,
        display_dict=bundle.display_dict,
    )


def plot_not_found(nChances, h_mark):
    pass


def plot_activity(bdict, figroot=None, results_db=None, out_dir=".", fig_format="pdf"):
    activity_deg = {}
    activity_days = {}
    for k in bdict:
        if "Chances_of_detecting_activity" in k:
            if "deg" in k:
                activity_deg[k] = bdict[k]
            if "days" in k:
                activity_days[k] = bdict[k]

    first = bdict[list(bdict.keys())[0]]
    if figroot is None:
        figroot = first.run_name
    display_dict = deepcopy(first.display_dict)

    if len(activity_days) > 0:
        # Plot (mean) likelihood of detection of activity over X days
        ph = plots.PlotHandler(fig_format=fig_format, results_db=results_db, out_dir=out_dir)
        ph.set_metric_bundles(activity_days)
        ph.joint_metric_names = "Chances of detecting activity lasting X days"
        plot_dict = {"ylabel": "Mean likelihood of detection", "figsize": (8, 6)}
        ph.plot(
            plot_func=plots.MetricVsH(),
            plot_dicts=plot_dict,
            display_dict=display_dict,
            outfile_root=figroot + "_activityDays",
        )
    if len(activity_deg) > 0:
        # Plot mean likelihood of detection of activity over X amount of orbit
        ph = plots.PlotHandler(fig_format=fig_format, results_db=results_db, out_dir=out_dir)
        ph.set_metric_bundles(activity_deg)
        ph.joint_metric_names = "Chances of detecting activity covering X deg"
        plot_dict = {"ylabel": "Mean likelihood of detection", "figsize": (8, 6)}
        ph.plot(
            plot_func=plots.MetricVsH(),
            plot_dicts=plot_dict,
            display_dict=display_dict,
            outfile_root=figroot + "_activityDeg",
        )


def read_and_combine(orbitRoot, baseDir, splits, metricfile):
    """Read and combine the metric results from split locations,
    returning a single bundle.

    This will read the files from
    baseDir/orbitRoot_[split]/metricfile
    where split = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9], etc.
    (the subsets the original orbit file was split into).

    Parameters
    ----------
    orbitRoot : `str`
        The root of the orbit file - l7_5k, mbas_5k, etc.
    baseDir: `str`
        The root directory containing the subset directories. (e.g. '.' often)
    splits:` np.ndarray` or `list` of `ints`
        The integers describing the split directories
        (e.g. [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    metricfile: `str`
        The metric filename.

    Returns
    -------
    metric_bundle : `~rubin_sim.maf.MoMetricBundle`
        A single metric bundle containing the combined data from the subsets.

    Note that this won't work for particularly complex metric values,
    such as the parent Discovery metrics. However, you can read and combine
    their child metrics, as for these we can propagate the data masks.
    """
    subsets = {}
    for i in splits:
        subsets[i] = mb.create_empty_mo_metric_bundle()
        ddir = os.path.join(baseDir, f"{orbitRoot}_{i}")
        subsets[i].read(os.path.join(ddir, metricfile))
    bundle = combine_subsets(subsets)
    return bundle


def combine_subsets(mbSubsets):
    # Combine the data from the subset metric bundles.
    # The first bundle will be used a template for the slicer.
    if isinstance(mbSubsets, dict):
        first = mbSubsets[list(mbSubsets.keys())[0]]
    else:
        first = mbSubsets[0]
        subsetdict = {}
        for i, b in enumerate(mbSubsets):
            subsetdict[i] = b
        mbSubsets = subsetdict
    joint = mb.create_empty_mo_metric_bundle()
    # Check if they're the same slicer.
    slicer = deepcopy(first.slicer)
    for i in mbSubsets:
        if np.any(slicer.slice_points["H"] != mbSubsets[i].slicer.slice_points["H"]):
            if np.any(slicer.slice_points["orbits"] != mbSubsets[i].slicer.slice_points["orbits"]):
                raise ValueError("Bundle %s has a different slicer than the first bundle" % (i))
    # Join metric values.
    joint.slicer = slicer
    joint.metric = first.metric
    # Don't just use the slicer shape to define the metric_values,
    # because of CompletenessBundles.
    metric_values = np.zeros(first.metric_values.shape, float)
    metric_values_mask = np.zeros(first.metric_values.shape, bool)
    for i in mbSubsets:
        metric_values += mbSubsets[i].metric_values.filled(0)
        metric_values_mask = np.where(metric_values_mask & mbSubsets[i].metric_values.mask, True, False)
    joint.metricValues = ma.MaskedArray(data=metric_values, mask=metric_values_mask, fill_value=0)
    joint.info_label = first.info_label
    joint.run_name = first.run_name
    joint.file_root = first.file_root.replace(".npz", "")
    joint.plotDict = first.plotDict
    joint.display_dict = first.display_dict
    return joint



import unittest

import numpy as np

from rubin_sim.moving_objects import chebeval, chebfit, make_cheb_matrix


class TestChebgrid(unittest.TestCase):
    def test_raise_error(self):
        x = np.linspace(-1, 1, 9)
        y = np.sin(x)
        dy = np.cos(x)
        p, resid, rms, maxresid = chebfit(x, y, dy, n_poly=4)
        with self.assertRaises(RuntimeError):
            chebeval(np.linspace(-1, 1, 17), p, interval=[1, 2, 3])

    def test_eval(self):
        x = np.linspace(-1, 1, 9)
        y = np.sin(x)
        dy = np.cos(x)
        p, resid, rms, maxresid = chebfit(x, y, dy, n_poly=4)
        yy_w_vel, vv = chebeval(np.linspace(-1, 1, 17), p)
        yy_wout_vel, vv = chebeval(np.linspace(-1, 1, 17), p, do_velocity=False)
        self.assertTrue(np.allclose(yy_wout_vel, yy_w_vel))
        # Test that we get a nan for a value outside the range of the
        # 'interval', if mask=True
        yy_w_vel, vv = chebeval(np.linspace(-2, 1, 17), p, mask=True)
        self.assertTrue(
            np.isnan(yy_w_vel[0]),
            msg="Expected NaN for masked/out of range value, but got %.2e" % (yy_w_vel[0]),
        )

    def test_ends_locked(self):
        x = np.linspace(-1, 1, 9)
        y = np.sin(x)
        dy = np.cos(x)
        for polynomial in range(4, 10):
            p, resid, rms, maxresid = chebfit(x, y, dy, n_poly=4)
            yy, vv = chebeval(np.linspace(-1, 1, 17), p)
            self.assertAlmostEqual(yy[0], y[0], places=13)
            self.assertAlmostEqual(yy[-1], y[-1], places=13)
            self.assertAlmostEqual(vv[0], dy[0], places=13)
            self.assertAlmostEqual(vv[-1], dy[-1], places=13)

    def test_accuracy(self):
        """If n_poly is  greater than number of values being fit,
        then fit should be exact."""
        x = np.linspace(0, np.pi, 9)
        y = np.sin(x)
        dy = np.cos(x)
        p, resid, rms, maxresid = chebfit(x, y, dy, n_poly=16)
        yy, vv = chebeval(x, p, interval=np.array([0, np.pi]))
        self.assertTrue(np.allclose(yy, y, rtol=1e-13))
        self.assertTrue(np.allclose(vv, dy, rtol=1e-13))
        self.assertLess(np.sum(resid), 1e-13)

    def test_accuracy_prefit_c1c2(self):
        """If n_poly is  greater than number of values being fit,
        then fit should be exact."""
        NPOINTS = 8
        NPOLY = 16
        x = np.linspace(0, np.pi, NPOINTS + 1)
        y = np.sin(x)
        dy = np.cos(x)
        xmatrix, dxmatrix = make_cheb_matrix(NPOINTS + 1, NPOLY)
        p, resid, rms, maxresid = chebfit(
            x, y, dy, x_multiplier=xmatrix, dx_multiplier=dxmatrix, n_poly=NPOLY
        )
        yy, vv = chebeval(x, p, interval=np.array([0, np.pi]))
        self.assertTrue(np.allclose(yy, y, rtol=1e-13))
        self.assertTrue(np.allclose(vv, dy, rtol=1e-13))
        self.assertLess(np.sum(resid), 1e-13)


if __name__ == "__main__":
    unittest.main()
